#include "as/AS.h"
#include "appcode.h"
#include "objects.h"

#include "photosensor.h"
#include "hw/dev/P7.h"
#include "hw/irq/DMA_SYSSRC2.h"
#include "stddef.h"

#define NUM_SENSORS 8

#include <tc1796b/scu.h>

static PHOTOSENSOR_callback_t l_errorCallback=NULL;
static char errorOnBelow=-1;
static char mask=0;
static char lastInterrupted=-1;
static bool risingEnabled=false;
static bool fallingEnabled=false;

void PHOTOSENSOR_Init() {
	for (int i = 0; i <= 7; i++) {
		hw::dev::P7::setMode(i, hw::dev::P7::INPUT);
	}

	{
		UInt16* eicr = (UInt16*) ( &EICR0.reg );
		*eicr =
			1 << EICRn_FEN0_SHIFT  |          // react on falling edge (1 = yes, 0 = no)
			0 << EICRn_REN0_SHIFT  |          // react on rising edge (1 = yes, 0 = no)
			0 << EICRn_LDEN0_SHIFT |          // automatic reset (level detection) (1 = yes, 0 = no)
			1 << EICRn_EIEN0_SHIFT |          // external interrupt enabled/disable
			0 << EICRn_INP0_SHIFT; // external interrupt node (0-3)
	}

	/* ISR aktivieren, SPRN wird durch CiAO gesetzt */
	hw::irq::DMA_SYSSRC2::enable();
}

void PHOTOSENSOR_EnableRisingEdge() { 
	UInt16* eicr = (UInt16*) ( &EICR0.reg );
	*eicr |= 1 << EICRn_REN0_SHIFT;

	risingEnabled = true;
}

void PHOTOSENSOR_DisableRisingEdge() {
	UInt16* eicr = (UInt16*) ( &EICR0.reg );
	*eicr &= ~(1 << EICRn_REN0_SHIFT);

	risingEnabled = false;
}

void PHOTOSENSOR_EnableFallingEdge() { fallingEnabled = true; }
void PHOTOSENSOR_DisableFallingEdge() { fallingEnabled = false; }

void PHOTOSENSOR_DisableBothEdges() {
	PHOTOSENSOR_DisableFallingEdge();
	PHOTOSENSOR_DisableRisingEdge();
}

bool PHOTOSENSOR_Status(char sensorno) {
	if ((sensorno < 0) || (sensorno >= NUM_SENSORS))
		return false;
	
	return !hw::dev::P7::get(sensorno);
}

char PHOTOSENSOR_NoInterrupted() {
	for(char i = 0; i <= 7; ++i) {
		if (! hw::dev::P7::get(i) )
			return i;
	}
	return -1;
}

void PHOTOSENSOR_MaskAll() { mask = 0x0; }
void PHOTOSENSOR_UnmaskAll() { mask = 0xff; }

void PHOTOSENSOR_Mask(char sensorno) {
	if((sensorno >= 0) && (sensorno <= 7)) { mask &= ~(1<<sensorno); }
}

void PHOTOSENSOR_MaskAllButOne(char sensorno) { mask = (1<<sensorno); }

void PHOTOSENSOR_Unmask(char sensorno) {
	if((sensorno >= 0) && (sensorno <= 7)) { mask |= (1<<sensorno); }
}


bool PHOTOSENSOR_IsMasked(char sensorno) {
	if((sensorno >= 0) && (sensorno <= 7)) {
		return (mask & (1 << sensorno)) == 0;
	} else {
		return true;
	}
}

void PHOTOSENSOR_SetErrorCallback(PHOTOSENSOR_callback_t func) { l_errorCallback = func; }
void PHOTOSENSOR_ErrorOnBelow(char sensorno) { errorOnBelow = sensorno; }

/*--------------------------------------------------------------------------*/
/* Category 2 interrupt: PSISR                                             */
/*--------------------------------------------------------------------------*/
void FSM::functionISRPSISR() {
	char interrupted = PHOTOSENSOR_NoInterrupted();
	bool risingEdge = (interrupted == -1);
	bool doCall = true;

	if(risingEdge) { // We got a rising Edge interrupt
		
		/* return if rising edges disabled */
		if((lastInterrupted == -1) || !risingEnabled) {
			doCall = false;
		}
		
	} else { // We got a falling Edge interrupt
		// Update lastInterrupted
		lastInterrupted = interrupted;
		
		/* return if this is a spurious falling edge at startup
		 * or falling edges disabled */
		if((lastInterrupted == -1) || !fallingEnabled) {
			doCall = false;
		}
	}
	
	if(lastInterrupted < errorOnBelow) {
		l_errorCallback(risingEdge, lastInterrupted);
		return;
	}
	
	/* If masked do not call callback */
	if(PHOTOSENSOR_IsMasked(lastInterrupted)) { doCall = false; }

	if(doCall) { AS::ActivateTask(FSMTask); }
}

