#include "appcode.h"
#include "as/AS.h"
#include "objects.h"

#include "fsm.h"
#include "device/photosensor.h"
#include "device/coil.h"
#include "control/sequence.h"
#include "control/command.h"
#include "common/timing.h"

/*! This enum defines the different states of the primary FSM */
typedef enum {
	FSM_FETCH,		// Fetching the next command
	FSM_UP,				// Executing the Up-Fsm
	FSM_DOWN,			// Executing the Down-Fsm
	FSM_HOLD,			// Executing the Hold-Fsm
	FSM_RELEASE,	// Executing the Release-Fsm
	FSM_END				// End state of the FSM
} state_fsm;


/*! This enum defines the substates of the Up-Fsm */
typedef enum {
	FSM_UP_START,
	FSM_UP_1,
	FSM_UP_2,
	FSM_UP_END
} substate_up;

/*! This enum defines the substates of the Down-Fsm */
typedef enum {
	FSM_DOWN_START,
	FSM_DOWN_1,
	FSM_DOWN_2,
	FSM_DOWN_END
} substate_down;

/*! This enum defines the substates of the Hold-Fsm */
typedef enum {
	FSM_HOLD_START,
	FSM_HOLD_END
} substate_hold;

/*! This enum defines the substates of the Release-Fsm */
typedef enum {
	FSM_RELEASE_START,
	FSM_RELEASE_END
} substate_release;

/*! The variables that save the current states */
static state_fsm state = FSM_FETCH;
static substate_up state_up = FSM_UP_START;
static substate_down state_down = FSM_DOWN_START;
static substate_hold state_hold = FSM_HOLD_START;
static substate_release state_release = FSM_RELEASE_START;

/*! This variable defines the number of the coil,
 *  where the projectile was short before or is now.
 */
static int currentCoil = 0;

/*! This variable counts the number of remaining coils of a real movement
 *  that have to be covered for the acutual command
 */
static int coilsToGo = 0;

/*! In some cases we get the amount of time to wait from the command
 *  In these cases we keep this amount here
 */
static int timeToWait = 0;

int FSM_Fetch() {
	/* If no overriden we want to wait indefinitly until next interrupt */
	int timeToNextStep = -1;
	
	/* Get the next command to be executed */
	command cmd = SEQUENCE_GetCommand();
	
	switch(cmd.mtype) {
		case MOVE_END:
			/* If finished, switch of all coils */
			COIL_AllOff();
			PHOTOSENSOR_ErrorOnBelow(8);
			break;
			
		case MOVE_UP:
			state = FSM_UP;
			state_up = FSM_UP_START;
			
			/* Here cmd.n has to be interpreted as the number of coils that have
			 * to be covered */
			coilsToGo = cmd.n;
			
			timeToNextStep = 0;
			break;
			
		case MOVE_DOWN:
			state = FSM_DOWN;
			state_down = FSM_DOWN_START;

			/* Here cmd.n has to be interpreted as the number of coils that have
			 * to be covered */
			coilsToGo = cmd.n;
			
			timeToNextStep = 0;
			break;
			
		case MOVE_HOLD:
			state = FSM_HOLD;
			state_hold = FSM_HOLD_START;
			
			/* Here cmd.n has to be interpreted as the duration of hold */
			timeToWait = cmd.n;
			
			timeToNextStep = 0;
			break;
			
		case MOVE_RELEASE:
			state = FSM_RELEASE;
			state_release = FSM_RELEASE_START;
			
			/* Here cmd.n has to be interpreted as the duration of release */
			timeToWait = cmd.n;
			
			timeToNextStep = 0;
			break;
			
		default:
			break;
	}
	
	return timeToNextStep;
}

/*! Execute one step of the Up-Fsm.
 *  This FSM requires that the projectile is 
 *  Below the actual currentCoil and in range of the field so that
 *  it can be accelerated to up to the next coil
 */
int FSM_Up() {
	int timeToNextStep = -1;
	
	switch(state_up) {
		case FSM_UP_START:
			/* Accelerate the projectile until photosensor is interrupted */
			PHOTOSENSOR_MaskAllButOne(currentCoil);
			PHOTOSENSOR_DisableBothEdges();
			PHOTOSENSOR_EnableFallingEdge();
			PHOTOSENSOR_ErrorOnBelow(currentCoil);
			COIL_On(currentCoil);
			
			state_up = FSM_UP_1;
			break;
			
		case FSM_UP_1:
			/* We are still in the currentCoil
			 * but now it is time to disable the coil.
			 * Disable it and enable the next coil.
			 * Wait until the next coil is reached
			 */
			PHOTOSENSOR_MaskAllButOne(currentCoil + 1);
			COIL_On(currentCoil + 1);
			PHOTOSENSOR_ErrorOnBelow(currentCoil);
			
			state_up = FSM_UP_2;
			break;
			
		case FSM_UP_2:
			/* We reached the upper next coil */
			currentCoil++;
			coilsToGo--;
			PHOTOSENSOR_ErrorOnBelow(currentCoil);
			
			if (coilsToGo == 0) {
				/* We have finished the up movement */
				state_up = FSM_UP_END;
				timeToNextStep = 0;
				PHOTOSENSOR_MaskAll();
			} else {
				/* Disable currentCoil and enable the next coil.
				 * Wait until the next coil is reached
				 */
				PHOTOSENSOR_MaskAllButOne(currentCoil + 1);
				PHOTOSENSOR_ErrorOnBelow(currentCoil);
				COIL_On(currentCoil + 1);
			}

			break;

		default:
			break;
	}
	
	return timeToNextStep;
}

/*! Execute one step of the Down-Fsm.
 *  This fsm required that the projectile is actually falling down to
 *  currentCoil - 1, but can also still pass currentCoil without reaction or
 *  being hold at currentCoil
 */
int FSM_Down()
{
	int timeToNextStep = -1;
	
	switch(state_down) {
		case FSM_DOWN_START:
			/* We await the signal that the down next sensor
			 * has been passed.
			 * Switch of currentCoil in case of previous hold cycle.
			 */
			PHOTOSENSOR_MaskAllButOne(currentCoil - 1);
			PHOTOSENSOR_DisableBothEdges();
			PHOTOSENSOR_EnableRisingEdge();
			PHOTOSENSOR_ErrorOnBelow(currentCoil - 1);
			
			COIL_Off(currentCoil);
			
			state_down = FSM_DOWN_1;
			break;
			
		case FSM_DOWN_1:
			/* This state defers the braking of the falling bullet to make the 
			 * the movement smoother or steps to the end of the Down-Fsm
			 */
			currentCoil--;
			coilsToGo--;
			
			PHOTOSENSOR_MaskAll();
			PHOTOSENSOR_DisableBothEdges();
			PHOTOSENSOR_ErrorOnBelow(currentCoil);
			
			if(coilsToGo == 0) {
				/* If reached the whiched Coil: finished */
				state_down = FSM_DOWN_END;
				timeToNextStep = 0;
			} else {
				/* Else: delay the braking by DELAY_TO_BRAKE.
				 * All coils are currently switched off */
				state_down = FSM_DOWN_2;
				timeToNextStep = DELAY_TO_BRAKE;
			}
			break;
			
		case FSM_DOWN_2:
			/* Now brake:
			 * Switch on currentCoil and and wait BRAKE_DURATION to
			 * slow down the falling of the projectile
			 */
			COIL_On(currentCoil);
			PHOTOSENSOR_ErrorOnBelow(currentCoil - 1);
			
			state_down = FSM_DOWN_START;
			timeToNextStep = BRAKE_DURATION;
			
			break;
			
		default:
			break;
	}
	
	return timeToNextStep;
}


/*! Execute one step of the Hold-Fsm. */
int FSM_Hold()
{
	int timeToNextStep = -1;

	switch(state_hold) {
		case FSM_HOLD_START:
			/* Switch on currentCoil to hold the projectile
			 * for timeToWait milliseconds. */
			PHOTOSENSOR_MaskAll();
			PHOTOSENSOR_DisableBothEdges();
			PHOTOSENSOR_ErrorOnBelow(currentCoil);
			COIL_On(currentCoil);
			
			timeToNextStep = timeToWait;
			
			state_hold = FSM_HOLD_END;
			break;
			
		default:
			break;
	}
	
	return timeToNextStep;
}

/*! Execute one step of the Release-Fsm. */
int FSM_Release()
{
	int timeToNextStep = -1;
	
	switch(state_release) {
		case FSM_RELEASE_START:
			/* Switch off currentCoil to let the the projectile fall down for
			 * for timeToWait milliseconds. */
			PHOTOSENSOR_MaskAll();
			PHOTOSENSOR_DisableBothEdges();
			PHOTOSENSOR_ErrorOnBelow(currentCoil);
			COIL_AllOff();
			
			timeToNextStep = timeToWait;
			
			state_release = FSM_RELEASE_END;
			break;
			
		default:
			break;
	}
	return timeToNextStep;
}


/*! The function executes exactly one step of the primary FSM */
int FSM_Step() {
	int timeToNextStep = -1;
	
	switch(state) {
		case FSM_FETCH:
			/* Fetch next command */
			timeToNextStep = FSM_Fetch();
			break;
			
		case FSM_UP:
			if (state_up == FSM_UP_END) {
				/* Up-Fsm has finished, immediately get next command */
				state = FSM_FETCH;
				timeToNextStep = 0;
			} else {
				/* Execute one step of the Up-Fsm */
				timeToNextStep = FSM_Up();
			}
			break;
			
		case FSM_DOWN:
			if (state_down == FSM_DOWN_END) {
				/* Down-Fsm has finished, immediately get next command */
				state = FSM_FETCH;
				timeToNextStep = 0;
			} else {
				/* Execute one step of the Down-Fsm */
				timeToNextStep = FSM_Down();
			}
			break;
			
		case FSM_HOLD:
			if (state_hold == FSM_HOLD_END) {
				/* Hold-Fsm has finished, immediately get next command */
				state = FSM_FETCH;
				timeToNextStep = 0;
			} else {
				/* Execute one step of the Hold-Fsm */
				timeToNextStep = FSM_Hold();
			}
			break;
			
		case FSM_RELEASE:
			if (state_release == FSM_RELEASE_END) {
				/* Release-Fsm has finished, immediately get next command */
				state = FSM_FETCH;
				timeToNextStep = 0;
			} else {
				/* Execute one step of the Release-Fsm */
				timeToNextStep = FSM_Release();
			}
			break;
			
		default:
			break;
	}
	
	return timeToNextStep;
}

