#include "control.h"
#include "common/common.h"
#include "program.h"
#include <stdio.h>
#include <common/timing.h>

bool CTRL_CheckAndTranslate(program_t* prog, config_t* targetConf) {
	prog_elem_t* actElem = prog->firstElem;
	int pos = 0;
	
	prog->errorPos = 0;
	
	while(actElem != NULL) {
		switch(actElem->type) {
			case up_continuous:
				if(!CTRL_UpContinuous(actElem->n, targetConf)) {
					fprintf(stderr, "ERROR UpContinuous(%d);\n", actElem->n);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    UpContinuous(%d);\n", actElem->n);
				break;
				
			case down_continuous:
				if(!CTRL_DownContinuous(actElem->n, targetConf)) {
					fprintf(stderr, "ERROR DownContinuous(%d);\n", actElem->n);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    DownContinuous(%d);\n", actElem->n);
				break;
				
			case osc_continuous:
				if(!CTRL_OscContinuous(actElem->n, actElem->rounds, targetConf)) {
					fprintf(stderr, "ERROR OscContinuous(%d, %d);\n", actElem->n, actElem->rounds);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    OscContinuous(%d, %d);\n", actElem->n, actElem->rounds);
				break;
				
			case up_stepwise:
				if(!CTRL_UpStepwise(actElem->n, targetConf)) {
					fprintf(stderr, "ERROR UpStepwise(%d);\n", actElem->n);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    UpStepwise(%d);\n", actElem->n);
				break;
				
			case down_stepwise:
				if(!CTRL_DownStepwise(actElem->n, targetConf)) {
					fprintf(stderr, "ERROR DownStepwise(%d);\n", actElem->n);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    DownStepwise(%d);\n", actElem->n);
				break;
				
			case osc_stepwise:
				if(!CTRL_OscStepwise(actElem->n, actElem->rounds, targetConf)) {
					fprintf(stderr, "ERROR OscStepwise(%d, %d);\n", actElem->n, actElem->rounds);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    OscStepwise(%d, %d);\n", actElem->n, actElem->rounds);
				break;
				
			case hold:
				if(!CTRL_Hold(actElem->n, targetConf)) {
					fprintf(stderr, "ERROR Hold(%d);\n", actElem->n);
					prog->errorPos = pos;
					return FALSE;
				}
				fprintf(stderr, "OK    Hold(%d);\n", actElem->n);
				break;
				
			case start:
			default:
				break;
		}
		
		pos++;
		actElem = actElem->next;
	}
	
	return TRUE;
}

bool CTRL_UpContinuous(int n, config_t* conf) {
	/* Check boundaries */
	if(((conf->currentCoil + n) >= NUM_COILS) || (n <= 0)) {
		return FALSE;
	}
	
	conf->currentCoil += n;
	
	/* Summarize commands of same type */
	switch(conf->lastElem->cmd.mtype) {
		case MOVE_UP:
			conf->lastElem->cmd.n += n;
			return TRUE;
			break;
		case MOVE_HOLD:
			CFG_AddRelease(conf, RELEASE_TIME);
			break;
		case MOVE_DOWN:
			CFG_AddRelease(conf, DOWN_RELEASE_TIME);
			break;
		default:
			break;
	}
	
	CFG_AddUp(conf, n);	
	return TRUE;
}

bool CTRL_UpStepwise(int n, config_t* conf) {
	int i = 0;
	
	/* Check boundaries */
	if(((conf->currentCoil + n) >= NUM_COILS) || (n <= 0)) {
		return FALSE;
	}
	
	/* Translate into Continuous snd hold commands */
	for(i = 0; i < n; ++i) {
		CTRL_UpContinuous(1, conf);
		CTRL_Hold(HOLD_DURATION, conf);
	}
	
	return TRUE;
}

bool CTRL_DownContinuous(int n, config_t* conf) {
	/* Check boundaries */
	if(((conf->currentCoil - n) < 0) || (n <= 0)) {
		return FALSE;
	} else {
		conf->currentCoil -= n;
	}
	
	/* Summarize commands of same type */
	switch(conf->lastElem->cmd.mtype) {
		case MOVE_DOWN:
			conf->lastElem->cmd.n += n;
			return TRUE;
			break;
		case MOVE_UP:
			CFG_AddHold(conf, UP_BRAKE_TIME);
			break;
		default:
			break;
	}
	
	CFG_AddDown(conf, n);
	return TRUE;
}

bool CTRL_DownStepwise(int n, config_t* conf) {
	int i = 0;
	
	/* Check boundaries */
	if(((conf->currentCoil - n) < 0) || (n <= 0)) {
		return FALSE;
	}
	
	/* Translate into Continuous commands */
	for(i = 0; i < n; ++i) {
		CTRL_DownContinuous(1, conf);
		CTRL_Hold(HOLD_DURATION, conf);
	}
	
	return TRUE;
}

bool CTRL_Hold(int n, config_t* conf) {
	/* Check boundaries */
	if((n > MAX_HOLD_TIME) || (n <= 0)) {
		return FALSE;
	}
	
	/* Summarize commands of same type */
	if(conf->lastElem->cmd.mtype == MOVE_HOLD) {
		/* MAX_HOLD_TIME must not be exceeded */
		if((conf->lastElem->cmd.n + n) > MAX_HOLD_TIME) {
			return FALSE;
		}
		conf->lastElem->cmd.n += n;
		return TRUE;
	}
	
	CFG_AddHold(conf, n);
	return TRUE;
}


bool CTRL_OscContinuous(int n, int rounds, config_t* conf) {
	if((n == 0) || (rounds <= 0) || ((conf->currentCoil + n) < 0)
			|| ((conf->currentCoil + n) >= NUM_COILS)) {
		return FALSE;
	}
	
	for (; rounds > 0; rounds--) {
		if (n > 0) {
			CTRL_UpContinuous(n, conf);
			CTRL_DownContinuous(n, conf);
		} else {
			CTRL_DownContinuous(-1 * n, conf);
			CTRL_UpContinuous(-1 * n, conf);
		}
	}
	
	return TRUE;
}

bool CTRL_OscStepwise(int n, int rounds, config_t* conf) {
	if((n == 0) || (rounds <= 0) || ((conf->currentCoil + n) < 0)
			|| ((conf->currentCoil + n) >= NUM_COILS)) {
		return FALSE;
	}
	
	for (; rounds > 0; rounds--) {
		if (n > 0) {
			CTRL_UpStepwise(n, conf);
			CTRL_DownStepwise(n, conf);
		} else {
			CTRL_DownStepwise(-1 * n, conf);
			CTRL_UpStepwise(-1 * n, conf);
		}
	}
	
	return TRUE;
}

config_t* CTRL_Compile(program_t* inProg) {
	config_t* conf = CFG_Create();
	
	if(!CTRL_CheckAndTranslate(inProg, conf)) {
		CFG_Delete(conf);
		return NULL;
	}
	
	if(conf->currentCoil > 0) {
		if(!CTRL_DownContinuous(conf->currentCoil, conf)) {
			CFG_Delete(conf);
			return NULL;
		}
	}
	
	if((conf->lastElem->cmd.mtype != MOVE_HOLD) 
			|| ((conf->lastElem->cmd.mtype == MOVE_HOLD) && (conf->lastElem->cmd.n < 300))) {
		if(!CTRL_Hold(300, conf)) {
			CFG_Delete(conf);
			return NULL;
		}
	}
	
	CFG_AddEnd(conf);
	
	return conf;
}
