/*
 * $Id: triboard1796_vectab.cpp 1019 2009-09-13 23:09:30Z sijnstre $
 * created:   07.03.2006 lohmann
 */  
 
/** 
 * @file triboard1796_vectab.cpp
 * @brief Generates (contains) TriCore IRQ and TRAP tables
 */ 

 

#include "triboard1796.h"
#include "stdinc.h"

// cfHAL_IRQHANDLER_ATTRIBUTES and cfHAL_TRAPHANDLER_ATTRIBUTES should be defined by p::v
#ifndef cfHAL_IRQHANDLER_ATTRIBUTES
	#warning cfHAL_IRQHANDLER_ATTRIBUTES was not defined by p::v  - defaulting to __attribute__((interrupt))
	#define cfHAL_IRQHANDLER_ATTRIBUTES __attribute__((interrupt))
#endif

#ifndef cfHAL_TRAPHANDLER_ATTRIBUTES
	#warning cfHAL_TRAPHANDLER_ATTRIBUTES was not defined by p::v - defaulting to __attribute__((interrupt))
	#define cfHAL_TRAPHANDLER_ATTRIBUTES __attribute__((interrupt))
#endif


#define IRQ_ENTRY_SECTION ".ciao_irqtab"
#define IRQ_ENTRY_SECTION_ATTRIBUTES __attribute__((asection(IRQ_ENTRY_SECTION, "a=32", "f=ax")))
#define IRQ_ENTRY_ATTRIBUTES __attribute__((interrupt_handler))\
                             IRQ_ENTRY_SECTION_ATTRIBUTES

#define IRQ_ENTRY_DEFHANDLER_ATTRIBUTES __attribute__((weak))\
										cfHAL_IRQHANDLER_ATTRIBUTES

#define TRAP_ENTRY_SECTION ".ciao_traptab"

#define TRAP_ENTRY_ATTRIBUTES __attribute__((interrupt_handler))\
                             __attribute__((asection(TRAP_ENTRY_SECTION, "a=32", "f=ax")))

#define TRAP_ENTRY_DEFHANDLER_ATTRIBUTES __attribute__((weak))\
										cfHAL_TRAPHANDLER_ATTRIBUTES

/* Generates an IRQ vector table entry. Entries are generated by placing an
 * aligned __ciao_tc_irq_<num>() function at an 32 Bytes aligned address into the
 * section .ciao_irqtab. Consider the IRQ name for IRQ 1 is ASC0_RCV, then the
 * generated function looks as:
 * 
 *   void __ciao_tc_irq_1() {
 *     asm("svlcx");
 *     hw::irq::ASC0_RCV::HWHandler();	// jump-and-link
 * 	   asm("rslcx");
 *     asm("rfe");
 *   }
 *  
 * The resulting code must not exceed  32 Bytes!
 *
 */   

// FIXME: "prologue" de-activating the call-depth counter can possibly be removed
// since this was probably only necessary with IRQ continuations
#define IRQ_ENTRY(num,name)\
extern "C" void IRQ_ENTRY_ATTRIBUTES __ciao_tc_irq_##num (){\
	asm volatile(\
		"svlcx\n"\
		"mov %d0, 0x0bff\n"\
		"mtcr $psw, %d0\n"\
		"isync\n"\
	);\
	hw::irq::name::HWHandler();\
	asm volatile(\
		"JL __ciao_tc_irq_end_handler\n"\
	);\
}

// FIXME: "prologue" de-activating the call-depth counter can possibly be removed
// since this was probably only necessary with IRQ continuations
#define IRQ_CAT2ENTRY(num,name,group)\
extern "C" void IRQ_ENTRY_ATTRIBUTES __ciao_tc_irq_##num (){\
	asm volatile(\
		"bisr " #group "\n"\
		"mov %d0, 0x0bff\n"\
		"mtcr $psw, %d0\n"\
		"isync\n"\
	);\
	hw::irq::name::HWHandler();\
	asm volatile(\
		"JL __ciao_tc_irq_end_handler\n"\
	);\
}

/*
uncomment for performance testing of non-continuation models
#undef IRQ_ENTRY
#define IRQ_ENTRY(num,name)\
extern "C" void IRQ_ENTRY_ATTRIBUTES __ciao_tc_irq_##num (){\
	asm volatile(\
		"svlcx\n"\
	);\
	hw::irq::name::HWHandler();\
}*/

#define IRQ_NOENTRY(num)\
extern "C" void IRQ_ENTRY_ATTRIBUTES __ciao_tc_irq_##num (){\
	_svlcx();\
	_debug();\
}	// rslcx; nop; rfe;


// Generates an entry for the TRAP vector table that invokes
// hw::trap::TRAP<num>::HWHandler(). Provides also a (weak) default definition for 
// hw::trap::TRAP<num>::HWHandler() that simply does a debug stop.
//
// The main difference to an IRQ handler is that d15 contains a trap 
// identification number, which is moved to d4 to be passed to the trap handler
//
// !! The resulting code must not exceed 8 Words (32 Bytes) !!
//
#define TRAP_ENTRY(num)\
namespace hw { namespace trap {\
	struct TRAP##num { static void cfHAL_TRAPHANDLER_ATTRIBUTES HWHandler(long); };\
}}\
extern "C" void TRAP_ENTRY_ATTRIBUTES __ciao_tc_trap_##num (){\
	_svlcx();\
	hw::trap::TRAP##num::HWHandler(_getD15());\
}\
namespace hw { namespace trap {\
	void TRAP_ENTRY_DEFHANDLER_ATTRIBUTES TRAP##num::HWHandler(long) {\
		_debug();\
	}\
}}


#ifndef __puma

// IRQ_ENTRY(0) is never used, as priority 0 is disables an SRN.
// Nevertheless, it occupies an entry in the vector table.
// We use this space here for the "end of handler" code
extern "C" void  __ciao_tc_irq_0() __attribute__((alias("__ciao_tc_irq_end_handler")));

#define IRQ_ENTRIES_BEGIN \
extern "C" void IRQ_ENTRY_SECTION_ATTRIBUTES __ciao_tc_irq_end_handler() {\
	asm volatile(\
		"jz %d2, 1f\n"\
		"rslcx\n"\
		"rfe\n"\
		"1:\n"\
		"rslcx\n"\
	);\
} //ret

#define IRQ_ENTRIES_END

// include p::v generated irqtable
#define __GENERATE_IRQTABLE__
#include "../cfIRQMap.h"
#undef __GENERATE_IRQTABLE__

// define the TRAP entry table
// TriCore defines 8 Trap "classes"
TRAP_ENTRY(0)	// MMU
namespace hw { namespace trap {
	struct TRAP1 { static void cfHAL_TRAPHANDLER_ATTRIBUTES HWHandler(long); };
}}
extern "C" void TRAP_ENTRY_ATTRIBUTES __ciao_tc_trap_1 (){
	_svlcx();
	hw::trap::TRAP1::HWHandler(_getD15());
}
TRAP_ENTRY(2)	// Instruction Error
//TRAP_ENTRY(3)	// Context Management
namespace hw { namespace trap {
	struct TRAP3 { static void cfHAL_TRAPHANDLER_ATTRIBUTES HWHandler(long); };
}}
extern "C" void TRAP_ENTRY_ATTRIBUTES __ciao_tc_trap_3 (){
	if( _getD15() == 4 )	// FATAL ERROR context list underflow
		_debug();
	_svlcx();
	hw::trap::TRAP3::HWHandler(_getD15());
}
namespace hw { namespace trap {\
	void TRAP_ENTRY_DEFHANDLER_ATTRIBUTES TRAP3::HWHandler(long) {
		_debug();
	}
}}
TRAP_ENTRY(4)	// System Bus and Peripheral Error
TRAP_ENTRY(5)	// Assertion
//TRAP_ENTRY(6)	// System Call
#include "hw/trap/TRAP6.h"
extern "C" void TRAP_ENTRY_ATTRIBUTES __ciao_tc_trap_6 (){
	_svlcx();
	hw::trap::TRAP6::HWHandler(_getD15());
}
TRAP_ENTRY(7)	// NMI

#endif	// __puma


// default implementation for UNUSED irqs
extern "C" cfHAL_IRQHANDLER_ATTRIBUTES void __ciao_tc_irq_UNUSED() {
	_debug();
}


