// $Id: ASC.cpp 1309 2010-11-17 20:33:43Z mike $
// Created:   25.03.2006 lohmann

/**
 * \file ASC.cpp
 * \brief TC1796 serial interfaces implementation
 *
 */

#include "ASC.h"
#include "hw/dev/CGU.h"
#include "hw/triboard1796.h"
#include "ciaoConfig.h"

#include <tc1796b/asc0.h>
#include <tc1796b/asc1.h>
#include <tc1796b/p5.h>


namespace hw { namespace dev {


void ASCBase::putc_( UInt8 c ) {
	while( !RTS() )
		;
	// clear RTS bit
	RTS( false );
	// write value
	send(c);
	
}
	
UInt8 ASCBase::getc_() {
	while( !RTR() )
		;
	RTR( false );
	return receive();
}

void ASCBase::setMode( UInt32 baud ) {
	
	asc_->CON.reg		= 0;
	
	/*  reload_value = fdv/512 * freq/16/baudrate -1  ==>
		reload_value = (512*freq)/(baudrate * 512*16) - 1
		fdv = (reload_value + 1) * (baudrate*512*16/freq)
		reload_value = (frequency / 32) / BAUDRATE - 1;
	*/
	unsigned frequency = CGU::getSysClockFreq();
	unsigned reload_value = (frequency / (baud * 16)) - 1;
	unsigned dfreq = frequency / (16*512);
	unsigned fdivider = (reload_value + 1) * (unsigned)baud / dfreq;
	
	// set baudrate
	asc_->BG.reg		= reload_value;
	asc_->FDV.reg		= fdivider;
	
	// set interface mode
	asc_->CON.bits.M 	= 1;	// 8 bit data, asynchronous
	asc_->CON.bits.REN	= 1;	// receiver enable
	asc_->CON.bits.FDE	= 1;	// fractional divider enable
//	asc_->CON.bits.FDE	= 0;	// one stop bit (default)

	// enable baud rate generator
	asc_->CON.bits.R = 1;
	
	RTS( true );
}


//! the one-and-only ASC0 instance
ASC0 ASC0::theASC0_;

void ASC0::init() {
	
	// set base address
	asc_ = (ASC_t*) ASC0_BASE;

	// Set P5.1 (TX line) to high
	// drive active and output mode
	P5_IOCR0.bits.PC1 	= 0x9;	// 1001 = Port Output Alternate 1 Function Push/Pull
	P5_OMR.bits.PS1 	= 1;	
		
	// enable ASC clock (for both ASC0 and ASC1)
	// also enables the whole ASC module!
	ASC0_CLC.bits.RMC	= 1;
	ASC0_CLC.bits.DISR	= 0;
	
	// initialize SRPNs
	// reorder barrier is needed so that the ASC module is enabled
	// before writing to its memory-mapped registers
	_reorder_barrier();
	IRQ_REC::init();
	IRQ_TRA::init();
	IRQ_ERR::init();
	IRQ_TBUF::init();
	
	// initialize interface with defaults
	setMode( cfDEV_ASC_BAUD );
	
}

//! the one-and-only ASC1 instance
ASC1 ASC1::theASC1_;

void ASC1::init() {
	
	// set base address
	asc_ = (ASC_t*) ASC1_BASE;

	// Set P5.3 (TX line) to high
	// drive active and output mode
	P5_IOCR0.bits.PC3 	= 0x9;	// 1001 = Port Output Alternate 1 Function Push/Pull
	P5_OMR.bits.PS3 	= 1;	
		
	// enable ASC clock (for both ASC0 and ASC1)
	// also enables the whole ASC module!
	ASC0_CLC.bits.RMC	= 1;
	ASC0_CLC.bits.DISR	= 0;
	
	// initialize SRPNs
	// reorder barrier is needed so that the ASC module is enabled
	// before writing to its memory-mapped registers
	_reorder_barrier();
	IRQ_REC::init();
	IRQ_TRA::init();
	IRQ_ERR::init();
	IRQ_TBUF::init();

	// initialize interface with defaults
	setMode( cfDEV_ASC_BAUD );
}

}} // namespace hw::dev
