#ifndef __HW_DEV_IDE_H__
#define __HW_DEV_IDE_H__

#include "stdinc.h"

#include "hw/dev/IOPort.h"
#include "hw/dev/PCI.h"

//#include <stdio.h>

/*
#define CLASSCODE          0x1
#define SUBCLASSCODE       0x1

#define PRIMARY_COMMAND_BLOCK     0x01F0
#define PRIMARY_CONTROL_BLOCK     0x03F4
#define SECONDARY_COMMAND_BLOCK   0x0170
#define SECONDARY_CONTROL_BLOCK   0x0374
*/

namespace hw {
namespace dev {

class IDE {

public:

	static const unsigned CLASSCODE = 0x1;
	static const unsigned SUBCLASSCODE = 0x1;

	// default I/O-addresses if bar[0-4] == 0 or -1
	static const unsigned PRIMARY_COMMAND_BLOCK   = 0x01F0;
	static const unsigned PRIMARY_CONTROL_BLOCK   = 0x03F4;
	static const unsigned SECONDARY_COMMAND_BLOCK = 0x0170;
	static const unsigned SECONDARY_CONTROL_BLOCK = 0x0374;

	// command description:
	// http://www.t13.org/Documents/UploadedDocuments/project/d1410r3b-ATA-ATAPI-6.pdf

	enum IDE_COMMAND {
		READ          = 0x20,
		READ_EXT      = 0x24,
		READ_DMA      = 0xC8,
		READ_DMA_EXT  = 0x25,
		WRITE         = 0x30,
		WRITE_EXT     = 0x34,
		READ_BUFFER   = 0xE4,
		IDENTIFY      = 0xEC
	};

	enum STATUS_BIT {
		ERROR,
		INDEX,
		CORR,
		DRQ,
		DSC,
		DWF,
		DRDY,
		BUSY
	};

	struct DeviceInfo {
		char serial[20 + 1];   // Serial Number (C-String)
		char firmware[8 + 1];  // Firmware Revision (C-String)
		char model[40 + 1];    // Model number (C-String)

		UInt16 cylinder_count;
		UInt16 head_count;
		UInt16 sectors_per_track;

		UInt64 sector_count;   // C * H * SPT
	};

	struct PhysicalRegionDescriptor {
		UInt32 address;
		UInt16 size;
		UInt16 flags;
	};

private:

	static IDE inst_;

	UInt8 vector_;

	DeviceInfo deviceinfo_;

	// TODO: Mehr Deskriptoren?
	//PhysicalRegionDescriptor prdt_[1];

	struct {
		IOPort data;
		IOPort error;
		IOPort sector_count;
		IOPort sector_number;
		IOPort cylinder_low;
		IOPort cylinder_high;
		IOPort head;
		IOPort command;

		void setBaseAddress(UInt16 baseaddr) {
			data.setAddress(baseaddr + 0);
			error.setAddress(baseaddr + 1);
			sector_count.setAddress(baseaddr + 2);
			sector_number.setAddress(baseaddr + 3);
			cylinder_low.setAddress(baseaddr + 4);
			cylinder_high.setAddress(baseaddr + 5);
			head.setAddress(baseaddr + 6);
			command.setAddress(baseaddr + 7);
		}
	} primary_command_, secondary_command_;

	struct {
		IOPort command;
		IOPort status;
		IOPort address;

		void setBaseAddress(UInt16 baseaddr) {
			command.setAddress(baseaddr + 0);
			status.setAddress(baseaddr + 2);
			address.setAddress(baseaddr + 4);
		}
	} busmaster_;

	struct {
		IOPort control;
		IOPort floppy;

		void setBaseAddress(UInt16 baseaddr) {
			control.setAddress(baseaddr+2);
			floppy.setAddress(baseaddr+3);
		}
	} primary_control_, secondary_control_;

	struct CHS {
		UInt16 cylinder;
		UInt8  head;
		UInt8 sector;
	};

	bool status_bit(UInt8 bit) {
		return primary_command_.command.inb() & (1 << bit);
	}

	void wait_for_status_bit(UInt8 bit, bool status) {
		// TODO: primary vs. secondary
		while ( static_cast<bool>(status_bit(bit)) != status) {
			// Nix
			//printf("Waiting: 0x%x...\n", primary_command_.command.inb());
			//for (volatile int i = 0; i < 100000000; ++i);
		}
	}

	void readblockN(void* data, int words = 256);
	void writeblock(const void* data);

	void setAddress(UInt8 sector, UInt16 cylinder, UInt8 head);
	//void setAddress28(UInt64 sector);

	//void setAddress48(UInt64 sector); // LBA48

	void updateDeviceInfo();

	UInt64 calcLBA(UInt16 cylinder, UInt8 head, UInt8 sector);

	/*
	LBA / spt = Result1 + Remainder1
	        S = Remainder1 + 1
	Result1 / heads = Result2 + Remainder2
	        C = Result2
	        H = Remainder2
	*/

	CHS calcCHS(unsigned long lba);

	void clearBuffer();

	//void setDMABuffer(void* address, UInt16 size);
	//void writePRD(bool enable, bool write);

public:

	static IDE& Inst() { return inst_; }

	void init_regs();
	void init();

	// reads 512-Byte block
	//void readblock(void* data, UInt8 sector, UInt16 cylinder, UInt8 head);  // CHS
	//void readblock(void* data, UInt64 sector);                              // LBA28, LBA48

	//void readblock48(void* data, UInt64 sector);                            // LBA48

	//void readblockdma48(void* data, UInt64 sector);                         // LBA48

	// writes 512-Byte block
	//void writeblock(const void* data, UInt8 sector, UInt16 cylinder, UInt8 head);
	//void writeblock(const void* data, UInt64 sector);

	//void writeblock48(const void* data, UInt64 sector);                     // LBA48

	const DeviceInfo& getDeviceInfo() { return deviceinfo_; }

	//UInt8 getDMAStatus();
};

}
}

#endif // __HW_DEV_IDE_H__
