/** \author Christian Meier
 *  \version $Id: HwObject.cpp 995 2009-08-11 14:26:31Z mike $
 */

#include "InterruptScheme.h"
#include "HwObject.h"
#include "Controller.h"
#include "HwObject_pcCfg.h"

namespace hw { namespace dev { namespace can { namespace multican {

void
HwObject::Init(HwObjectConfigType const * p_Config) {
	// Although most of the stuff will only be initialized to the reset value, it is a good idea to list
	// almost all bitfield elements here to not forget anything.
	// Moreover some errata workarounds are made here.

	/* MOFCR:
	 * mofcr.bits.MMC  = 0;		We currently only use standard HwObjects, not FIFO or Gateway --> MMC = 0000b
	 * mofcr.bits.GDFS = 0;		Unused for us, write init value
	 * mofcr.bits.IDC	= 0;	Unused for us, write init value
	 * mofcr.bits.DLCC	= 0;	Unused for us, write init value
	 * mofcr.bits.DATC	= 0;	Unused for us, write init value
	 * mofcr.bits.RXIE = 0;		Here we clear the rx interrupt enable flag; depending on config
	 *							it will be reenabled somewhere else.
	 *	mofcr.bits.TXIE = 0;	Here we clear the tx interrupt enable flag; depending on config
	 *							it will be reenabled somewhere else.
	 *	mofcr.bits.OVIE = 0;	Unused for us, write init value
	 *	mofcr.bits.FRREN = 1;	Workaround Errata MultiCAN_TC.032, MultiCAN_AI.040
	 *	mofcr.bits.RMM 	= 0;	Unused for us, write init value
	 *	mofcr.bits.SDT	= 1;	RX: We don't want to use a shadow buffer (this is possible because tc is little endian)
	 *							TX: OK because we don't want to use remote frames.
	 *	mofcr.bits.STT	= 0;	We want to have automatic retransmit: Single Transmit Trial = 0
	 *	mofcr.bits.DLC	= 0;	He we initialize DLC to zero, it will be updated with every tx request and successful receive
	 *  mofcr.bits.MSC	= 0;	Only for TTCAN, write init value
	 */
	_MOFCR.reg = ((1 << CAN_MOFCRm_FRREN_SHIFT) & CAN_MOFCRm_FRREN_MASK) |
				 ((1 << CAN_MOFCRm_SDT_SHIFT) & CAN_MOFCRm_SDT_MASK);

	/* MOFGPR:
	 * Some error workarounds must be made here
	 * Errata:	MultiCAN_TC.032
	 *			MultiCAN_TC.028
	 *			MultiCAN_AI.040
	 *			MultiCAN_TC.025
	 */
	_MOFGPR.bits.CUR = Id();

	{ // MOIPR
		CAN_MOIPRm_t_nonv moipr;
		moipr.reg = 0;

		// Scheme: MPN[7:0] = HwObject handle
		moipr.bits.MPN = Id();

		// Scheme: INP[3:2] = ControllerId, INP[1] = 0 (unused), INP[0] = 0 (rx) / 1(tx)
		moipr.bits.TXINP = HW_DEV_CAN_MULTICAN_TX_INT(p_Config->m_ControllerId);
		moipr.bits.RXINP = HW_DEV_CAN_MULTICAN_RX_INT(p_Config->m_ControllerId);

		// Unused for us, write init value
		moipr.bits.CFCVAL	= 0;

		// Commit to hardware
		_MOIPR.reg = moipr.reg;
	}

	/* MOAMR:
	 * _MOAMR.bits.MIDE = 0;		For now we just allow either standard or extended ids per HwObject
	 * _MOAMR.bits.AM	= 0;		Here we clear the acceptance mask; depending on config it will be changed later.
	 */
	_MOAMR.reg = 0;

	/* MODATAH + MODATAL:
	 * No need to initialize Data registers
	 */

	/* MOAR
	 * We currently only use the can id priority oriented transmit
	 * arbitration PRI=10b and only one reception priority class, so
	 * we can initialize _MOAR.PRI to 2 = 10b priority class for all
	 * receive and transmit HwObjects
	 *
	 * _MOAR.bits.PRI = 2; // MultiCAN_TC.023
	 * _MOAR.bits.IDE = cfHW_DEV_CAN_TC_MULTICAN_HWOBJECT_DEFAULT_IDE;
	 */
	_MOAR.reg =
		((2 << CAN_MOARm_PRI_SHIFT) & CAN_MOARm_PRI_MASK) |
		((cfHW_DEV_CAN_TC_MULTICAN_HWOBJECT_DEFAULT_IDE << CAN_MOARm_IDE_SHIFT) & CAN_MOARm_IDE_MASK);

	{ // MOCTR (Attention: This register is write only, so any read-write cycle won't do what you expect.
		CAN_MOCTRm_new_t_nonv moctr;
		moctr.reg = 0;					// Requests initialized to 0

		moctr.bits.SETTXEN1 = 1;		// We do not use FIFOs --> we can permanently set TXEN1 to 1
										// Independently of later being configured as rx or tx HwObject.

		// CHECK: Maybe we should reset all other flags

		// Commit to hardware
		_MOCTR.reg = moctr.reg;
	}

	// Errata Workarounds: We only use list allocation at initialization time before configuring/enabling
	// Controllers. MultiCAN_AI.043, MultiCAN_TC.031

	// At last we assign the HwObject to the correct controller
	ListManager::ListManager::Inst().StaticAllocate(
			Controller::CtrlIdToCtrlUnit(p_Config->m_ControllerId) + 1,
			Id());
	// Aspects bind here (after)
}

}}}}
