// $Id: Continuation.h 995 2009-08-11 14:26:31Z mike $
// Created:   21.05.2006 lohmann

/**
 * @file Continuation.h
 * @brief Basic threading functionality
 *
 * This file declares the Continuation class, which implements the fundamental concept of a thread of control
 * (a Continuation).
 */


#ifndef __HW_HAL_CONTINUATION_H__
#define __HW_HAL_CONTINUATION_H__

#include "stdinc.h"
#include "x86Context.h"

// !Continuation start functions must obey the following (architecure-dependend) attributes
#define cfHAL_STARTFUNC_ATTRIBUTES		// nothing special on tricore

namespace hw { namespace hal {


class Continuation;

extern "C" {
	void Continuation_go(x86Context*);
	void Continuation_switch (x86Context* save_to, x86Context* load_from);
}

/**
* @brief Fundamental parts of class Continuation
*
* ContinuationBase contains the critical "sealed" elements of Continuation and should
* not be adviced by aspects. Aspect should extend class Continuation instead.
*
* ContinuationBase implements the elementary task switching operations, which
* are save(), load() and go()
*/
class ContinuationBase {
  protected:
	x86Context			context_;

  public:

	//! Type of a Continuation starter function
	typedef void (* cfHAL_STARTFUNC_ATTRIBUTES StartFunc) (Continuation* me);

  protected:

	//! Initializes the Continuation
	CIAO_STD void init() {
		context_.clear();
	}

	//! Starts a new continuation control flow in the passed function using tos as the top of stack.
	//  Never returns, the current thread's context is not saved!
	CIAO_STD void CIAO_INLINE  go( void* tos, StartFunc starter ) {

		// new continuations always start with no context
        context_.ebx = 0;
        context_.esi = 0;
        context_.edi = 0;
        context_.ebp = 0;

        void** ptos = (void**) tos;

        *(--ptos) = this;
        *(--ptos) = (void*) 0xdeadbeef;

        // Kickoff
        *(--ptos) = (void*) starter;
        context_.esp = ptos;

        Continuation_go(&context_);
	}

  public:
	  x86Context* getContext() { return &context_; }

};

/**
 * @brief Provides the basic thread abstraction of CiAO, which is a Continuation.
 *
 * Continuation provides the basic context change operation: start(), switchto(), saveAndStart(), saveAndSwitchto().
 * It provides furthermore means for aspects  to hook into the context change. As for all functions related
 * to context changes the resulting assembler code is critical, aspects should bind to specific hook
 * methods only: before_CPURelease(), before_LastCPURelease(), after_CPUReceive(), and after_FirstCPUReceive()
 * are guaranteed to be invoked by all context changing methods.
 */
class Continuation : public ContinuationBase {

	void internal_saveAndStart( void* tos, StartFunc starter, Continuation* to );
	void internal_saveAndSwitchto( Continuation* to );

	static void start_impl( Continuation* obj, void* tos, StartFunc starter, Continuation* to );
	void saveAndStart_impl( void* tos, StartFunc starter, Continuation* to );
	static void switchto_impl( Continuation* obj, Continuation* to );
	void saveAndSwitchto_impl( Continuation* to );

 public:
	typedef ContinuationBase Super;

	/**
	 * @brief Starts a new continuation control flow in the passed
	 * \p starter function using \p tos as the top of stack and \p to as context instance.
	 *
	 * @note
	 * - start() never returns. Therefore all call-stack related resources should have
	 *   been freed when start() is invoked.
	 * - start() triggers before_LastCPURelease(), the current thread's context is <b>not saved</b>.
	 * - The passed \p starter is supposed to invoke after_FirstCPUReceive().
	 */
	CIAO_STD void CIAO_INLINE start( void* tos, StartFunc starter, Continuation* to ) {
		hw::JUMP4( start_impl, this, tos, starter, to );
		before_LastCPURelease( to );
		to->go( tos, starter );

		// we will never come here
	}

	/**
	 * @brief Starts a new continuation control flow in the passed
	 * \p starter function using \p tos as the top of stack and \p to as context instance.
	 *
	 * @note
	 * - saveAndStart() triggers before_CPURelease() and (when getting the CPU again) after_CPUReceive()
	 * - The passed \p starter is expected to invoke after_FirstCPUReceive().
	 */
	CIAO_STD void CIAO_INLINE saveAndStart( void* tos, StartFunc starter, Continuation* to ) {
		saveAndStart_impl( tos, starter, to );
		after_CPUReceive();
	}

	/**
	 * @brief Switches to another Continuation control flow without saving own context.
	 *
	 * @note
	 * - switchto() never returns. Therefore all call-stack related resources should have
	 *   been freed when switchto() is invoked.
	 * - switchto() triggers before_LastCPURelease()
	 * - The behaviour is undefinied, if /to has never run before (if no saved context does exist).
	 */
	CIAO_STD void CIAO_INLINE switchto( Continuation* to ) {
		hw::JUMP2( switchto_impl, this, to );
	}

	/**
	 * @brief Switches to another Continuation control flow.
	 *
	 * @note
	 * - saveAndSwitchto() triggers before_CPURelease() and (when getting the CPU again) after_CPUReceive()
	 * - The behaviour is undefinied, if /to has never run before (if no saved context does exist).
	 */
	CIAO_STD void CIAO_INLINE saveAndSwitchto( Continuation* to ){
		saveAndSwitchto_impl( to );
		internal_saveAndSwitchto( to );
		after_CPUReceive();
	}

	/**
	 * @brief Hook function, to be execution-adviced by aspects that are interested in context changes.
	 *
	 * before_CPURelease() is called whenever a context change operation to another continuation is about to
	 * take place. It is called in the context of the thread that releases the CPU before its context
	 * saved. The continuation to switch to is passed as reference parameter and may be altered.
	 *
	 * Note that before_LastCPURelease() is triggered instead if the calling contination context is
	 * not saved, thus, will never continue.
	 */
	CIAO_STD void CIAO_INLINE before_CPURelease( Continuation*& to ) {

	}

	CIAO_STD void CIAO_INLINE before_LastCPURelease( Continuation*& to ) {

	}

	/**
	 * @brief Hook function, to be execution-adviced by aspects that are interested in context changes.
	 *
	 * after_CPUReceive() is called whenever a context change operation to another continuation has taken place.
	 * It is called in the context of the thread that just received the CPU.
	 *
	 * after_FirstCPUReceived() is triggered instead if a thread receives the CPU for the very first time.
	 * after_FirstCPUReceived() has to be called by the thread start function (e.g. kickoff()).
	 *
	 * @note after_CPUReceived() and after_FirstCPUReceived() invoke unlock() to end a potentially
	 * protected context change operation. Hence, before-advice should be used if the additional
	 * behavior has to be protected as well.
	 */
	CIAO_STD void CIAO_INLINE after_CPUReceive() {
		unlock();
	}
	CIAO_STD void CIAO_INLINE after_FirstCPUReceive() {
		unlock();
	}

	/**
	 * @brief Locks the dispatcher. Actual implementation is empty, but may be changed by advice.
	 *
	 * If context changes have to be atomic, lock() should be called before the
	 * the operation that actually performs the context switch. In every case (independed of
	 * lock() having been called or not) unlock() is invoked after the context switch.
	 *
	 * In the default implementation, lock() is an empty method. Locking functionality has
	 * to be applied by a locking aspect.
	 *
	 * @note Both, lock() and unlock() are idempotent operations. Invoking lock() on the already
	 * locked dispatcher has no effect, as invoking unlock() on the not-locked dispatcher.
	 */
	 CIAO_STD static void CIAO_INLINE lock() {
	 	// to be adviced
	 }

	/**
	 * @brief Unlocks the dispatcher. Actual implementation is empty, but may be changed by advice.
	 *
	 * If context changes have to be atomic, lock() should be called before the
	 * the operation that actually performs the context switch.  In every case (independed of
	 * lock() having been called or not) unlock() is invoked after the context switch from
	 * after_CPUReceive() and after_FirstCPUReceive().
	 *
	 * In the default implementation, unlock() is an empty method. Locking functionality has
	 * to be applied by a locking aspect.
	 *
	 * @note Both, lock() and unlock() are idempotent operations. Invoking lock() on the already
	 * locked dispatcher has no effect, as invoking unlock() on the not-locked dispatcher.
	 */
	 CIAO_STD static void CIAO_INLINE unlock() {
	 	// to be adviced
	 }



	/**
	 * @brief Retuns true, if the Continuation has at least once released the CPU and thereby
	 *  may be activated with switchto().
	 */
	CIAO_STD bool hasSeenCPU() {
		return context_.esp;
	}

};

} } // namespace hw::hal


#endif // __HW_HAL_CONTINUATION_H__
