// $Id: SchedStrategy.h 1265 2010-09-07 13:47:31Z mike $
// Created:   27.06.2006 lohmann

#ifndef __OS_SCHEDSTRATEGY_STATIC_H__
#define __OS_SCHEDSTRATEGY_STATIC_H__

#include "Task.h"
#include "TasklistTypes.h"
#include "hw/hal/TCBUser.h"

namespace os { namespace krn {


	class SchedStrategy {
		public:

			typedef TasklistTypes::Tasklist< 
				TaskType,
				Task::Pri,
				Task::MAX,
				Task::PRI_MAX,
				Task::PER_PRI_MAX >		Tasklist;
			typedef hw::hal::TCBUser	TCB;

			typedef TaskType			Id;


		private:

			Tasklist 		ready_;
			TaskType		running_;
		
			/*
			 * Tracks the task in the protection context of which the current continuation
			 * executes. This only differs from running_ when the system is idle. We need
			 * it in non-trusted ISRs to determine the stack bounds of the currently used
			 * stack.
			 */	
			TaskType		activeContext_;

			//volatile at least necessary for idle loop
			volatile UInt32    need_reschedule_; // set to int so the pcp can handle this correctly	
			// The thread control blocks	
			mutable TCB  tcbs_[ Task::MAX ];


			// To be adviced
			void idleLoop();
			void idle();

		protected:	
		public:		// TODO: Should be protected, however, ac++ has problems with that 

			// To be adviced by synchronisation aspect
			void lock() {
			}
			void unlock() {
			}


			// SchedImpl interface

			/**
			 * Picks the next task to run and returns it. 
			 * Always returns a task and clears needReschedule.
			 * 
			 * If prefer_next is false, nexttask is taken only from the ready list 
			 * if its priority is higher, otherwise the running (calling) task is returned.
			 * 
			 * If prefer_next is true, the returned task is always taken from the ready list unless
			 * the ready list is empty. In such case, the calling task runs idleLoop() until
			 * a task gets ready. The task which eventually gets ready might be the calling task itself, 
			 * hence getRunning() != schedule( true ) is not guaranteed.
			 */
			TaskType schedule( Bool prefer_next );

			/**
			 * Behaves as 'schedule' with 'prefer_next == false', but 
			 * also returns a task with the same priority -- not only
			 * tasks with higher priority.
			 * This behavior is necessary to implement a round round
			 * or FIFO schedule.
			 */
			TaskType takeNext();

			void setReady( TaskType whom );

			void setActive( TaskType whom ) {
				setReady( whom );
			}

			void setWaiting( TaskType whom ) {
				Task::setState( whom, TASK_WAITING );
				//ready_.remove( whom, Task::getPri( whom ));
			}

			void setSuspended( TaskType whom ){
				Task::setState( whom, TASK_SUSPENDED );
			}

			void setInterrupted( TaskType whom, Bool val ){
				Task::setInterrupted( whom, val );
			}

			void setRunning( TaskType whom );

			TCB* allocTCB( TaskType newTask ) const {
				TCB* tcb = &tcbs_[ newTask ];
				tcb->init( Task::getFunc( newTask ), newTask );
				return tcb;
			}

			Ptr allocStack( TaskType newTask ) const {
				return Task::getStack( newTask );
			}

		public:

			TCB* getTCB( TaskType of ) const {
				return &tcbs_[ of ];
			}
			Bool isReady( TaskType whom ) const {
				return Task::getState( whom ) == TASK_READY;
			}
			Bool isWaiting( TaskType whom ) const {
				return Task::getState( whom ) == TASK_WAITING;
			}
			Bool isSuspended( TaskType whom ) const {
				return Task::getState( whom ) == TASK_SUSPENDED;
			}
			Bool isInterrupted( TaskType whom ) const {
				return Task::getInterrupted( whom );
			}
			TaskType getRunning() const {
				return running_;
			}
			TaskType getActiveContext() const {
				return activeContext_;
			}
			void setNeedReschedule() {
				// sirupfis 14.02.2008
				// need_reschedule_ = true;
				need_reschedule_ = 1;
			}
			void clearNeedReschedule() {
				// sirupfis 14.02.2008
				// need_reschedule_ = false;
				need_reschedule_ = 0;
			}
			Bool needReschedule() const {
				// sirupfis 14.02.2008
				// return need_reschedule_;
				return (need_reschedule_ != 0);
			}

			void setTaskTable( UInt size = Task::MAX );
	};

}} // namespace os::krn

#endif // __OS_SCHEDSTRATEGY_STATIC_H__
