// $Id: compiler.h 995 2009-08-11 14:26:31Z mike $
// Created:   21.05.2006 lohmann

/**
 * @file compiler.h
 * @brief contains compiler (tricore-g++) specific definition macros for commonly used attributes and pragmas
 *
 * The tricore-g++ compiler supports some special function and type-level attributes that can be quite useful
 * to generate more compact or more optimal code.
 */

#ifndef __COMPILER_H__
#define __COMPILER_H__

//! Always inline this function
#define CIAO_INLINE __attribute__((always_inline))

//! Never inline this function
#define CIAO_NOINLINE __attribute__((noinline))

//! Symbol should have weak linkage
#define CIAO_WEAK __attribute__((weak))

//! Symbol should be an alias for symbol target
#define CIAO_ALIAS(target) __attribute__((alias(#target)))

//! Symbol should be an weak alias for symbol \a target
#define CIAO_WEAKALIAS(target) __attribute__((weak, alias(#target)))

//! Global symbol (function, method, variable) should be named \a name in the generated assembler output.
#define CIAO_NAME(name) asm(#name)

//! Align element to \a byte boundary
 #define CIAO_ALIGNED(byte) __attribute__((aligned(byte)))

//! Dummy
#define CIAO_USE_JL

//! Dummy
#define CIAO_USE_J_RFE

//! Assume that at call to this function will never return.
#define CIAO_NORETURN __attribute__((noreturn))


namespace hw {

#ifndef __puma

	/**
	 * @name JUMP generators
	 * @brief J(ump) to a function instead of calling it.
	 *
	 * The function address is casted to a function pointer that uses
	 * __attribute__((interrupt_handler)). If a function with this attribute is called, tricore-gcc generates
	 * a J <function> instead of a CALL <function>. As we are still in C++ land, it is however guaranteed
	 * that all parameters end up in the correct registers :-)
	 *
	 * @warning Unfortunately ac++ tends to crash when parsing these templates, hence they are
	 * conditionally defined using #ifdef __puma. Note that this can cause them to be removed
	 * completly if they are used in slices!
	 *
	 */
	//@{
	inline void CIAO_INLINE JUMP( void (*func)() ) {
		typedef void (* CIAO_USE_J_RFE Jump)();
		( (Jump)func )();
	}
	template< class A1 > inline void CIAO_INLINE JUMP1( void (*func)( A1 ), A1 a1) {
		typedef void (* CIAO_USE_J_RFE Jump1)( A1 );
		( (Jump1)func )( a1 );
	}
	template< class A1, class A2 > inline void CIAO_INLINE JUMP2( void (*func)( A1, A2 ), A1 a1, A2 a2 ) {
		typedef void (* CIAO_USE_J_RFE Jump2)( A1, A2 );
		( (Jump2)func )( a1, a2 );
	}
	template< class A1, class A2, class A3 > inline void CIAO_INLINE JUMP3( void (*func)( A1, A2, A3), A1 a1, A2 a2, A3 a3 ) {
		typedef void (* CIAO_USE_J_RFE Jump3)( A1, A2, A3);
		( (Jump3)func )( a1, a2, a3 );
	}
	template< class A1, class A2, class A3, class A4 > inline void CIAO_INLINE JUMP4( void (*func)( A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4 ) {
		typedef void (* CIAO_USE_J_RFE Jump4)( A1, A2, A3, A4);
		( (Jump4)func )( a1, a2, a3, a4 );
	}
	//@}
#else
	inline void JUMP(...) {}
	inline void JUMP1(...) {}
	inline void JUMP2(...) {}
	inline void JUMP3(...) {}
	inline void JUMP4(...) {}

#endif

}	// namespace hw


 #endif // __COMPILER_H__
