#ifndef UTIL_BITSET_MULTI_ELEM_IMPL_H
#define UTIL_BITSET_MULTI_ELEM_IMPL_H

#include "util/metaprog.h"
#include "util/bitmanip.h"
#include "util/types.h"

template<class ELEM_TYPE, int NUM_ELEMS>
class BitsetMultiElemImplBase {
protected:
	typedef typename uint_t_value_opt<NUM_ELEMS>::T ArrayIndexType;
	typedef BitsetMultiElemImplBase<ELEM_TYPE,NUM_ELEMS> BitsetType;

private:
	typedef ELEM_TYPE	ElemType;

	enum { BITS_PER_ELEM			= 8*sizeof(ElemType) };
	enum { BIT_TO_ARRAY_SHIFT	= metaprog::log2_bottom<BITS_PER_ELEM>::RES };

public:
	typedef typename uint_t_value_opt<BITS_PER_ELEM * NUM_ELEMS + 1>::T BitIndexType;

	inline void
	Set(BitIndexType BitIndex) {
		register ArrayIndexType ArrayIndex   = BitIndexToArrayIndex(BitIndex);
		register BitIndexType BitInElemIndex = BitIndex - ArrayIndex * BITS_PER_ELEM;

		ElemType & Elem = RawSet_[ArrayIndex];
		Elem = bitmanip::setBit(Elem, BitInElemIndex);
	}

	inline void
	Insert(BitIndexType BitIndex, UInt NewState) {
		register ArrayIndexType ArrayIndex   = BitIndexToArrayIndex(BitIndex);
		register BitIndexType BitInElemIndex = BitIndex - ArrayIndex * BITS_PER_ELEM;

		ElemType & Elem = RawSet_[ArrayIndex];
		Elem = bitmanip::insertBit(Elem, NewState, BitInElemIndex);
	}

	inline void
	Clear(BitIndexType BitIndex) {
		register ArrayIndexType ArrayIndex   = BitIndexToArrayIndex(BitIndex);
		register BitIndexType BitInElemIndex = BitIndex - ArrayIndex * BITS_PER_ELEM;

		ElemType & Elem = RawSet_[ArrayIndex];
		Elem = bitmanip::clearBit(Elem, BitInElemIndex);
	}

	inline Bool
	Test(BitIndexType BitIndex) const {
		register ArrayIndexType ArrayIndex   = BitIndexToArrayIndex(BitIndex);
		register BitIndexType BitInElemIndex = BitIndex - ArrayIndex * BITS_PER_ELEM;

		return bitmanip::testBit(RawSet_[ArrayIndex], BitInElemIndex);
	}

	inline BitIndexType
	Lowest1Bit() const {
		register ArrayIndexType FirstNotZeroIndex = GetFirstNotZeroElem();
		return bitmanip::lowest1Bit(RawSet_[FirstNotZeroIndex]) + FirstNotZeroIndex * BITS_PER_ELEM;
	}

	inline BitIndexType
	Lowest0Bit() const {
		register ArrayIndexType FirstNotAllSetIndex = GetFirstNotAllSetElem();
		return bitmanip::lowest0Bit(RawSet_[FirstNotAllSetIndex]) + FirstNotAllSetIndex * BITS_PER_ELEM;
	}

	inline BitIndexType
	Highest1Bit() const {
		register ArrayIndexType LastNotZeroIndex = GetLastNotZeroElem();
		return bitmanip::highest1Bit(RawSet_[LastNotZeroIndex]) + LastNotZeroIndex * BITS_PER_ELEM;
	}

	inline BitIndexType
	Highest0Bit() const {
		register ArrayIndexType LastNotAllSetIndex = GetLastNotAllSetElem();
		return bitmanip::highest0Bit(RawSet_[LastNotAllSetIndex]) + LastNotAllSetIndex * BITS_PER_ELEM;
	}

	void
	ClearAll() {
		// FIXME For now simple impl, later maybe metaprog
		ArrayIndexType i = 0;
		// ATTENTION: As NUM_ELEMS can always be the maximum for ArrayIndexType,
		// we always stop at the second last iteration.
		for( ; i < NUM_ELEMS - 1; ++i) {
			RawSet_[i] = 0;
		}
		RawSet_[i] = 0;
	}

	Bool
	AllCleared() const {
		// FIXME For now simple impl, later maybe metaprog
		ArrayIndexType i = 0;
		ElemType OredAcc = 0;
		// ATTENTION: As NUM_ELEMS can always be the maximum for ArrayIndexType,
		// we always stop at the second last iteration.
		for( ; i < NUM_ELEMS - 1; ++i) {
			OredAcc |= RawSet_[i];
		}
		OredAcc |= RawSet_[i];
		return OredAcc == (ElemType)0;
	}

  class IndexIteratorOnes {
  public:
    inline IndexIteratorOnes()
      : MyBitset_(0),
				MyBitIndex_(NUM_ELEMS*BITS_PER_ELEM) {}

    inline IndexIteratorOnes(BitsetType * MyBitset)
      : MyBitset_(MyBitset),
				MyBitIndex_(0) {}

		__attribute__((noinline)) Bool
		GetNext(BitIndexType & FoundIndex) {
			if(MyBitIndex_ >= NUM_ELEMS*BITS_PER_ELEM) {
				return false;
			}

			register ArrayIndexType ArrayIndex   = BitIndexToArrayIndex(MyBitIndex_);
			register BitIndexType BitInElemIndex = MyBitIndex_ - ArrayIndex * BITS_PER_ELEM;

			register ElemType ActElem =
				MyBitset_->RawSet_[ArrayIndex] & (metaprog::TwoPowI_minus1<BITS_PER_ELEM>::RES << BitInElemIndex);

			do {
				if(ActElem) {
					break;
				}
				if(ArrayIndex >= (NUM_ELEMS - 1)) {
					break;
				}
				ActElem = MyBitset_->RawSet_[++ArrayIndex];
			} while(1);

			if(ActElem) {
				BitInElemIndex = bitmanip::lowest1Bit(ActElem);
				FoundIndex = ArrayIndex*BITS_PER_ELEM + BitInElemIndex;

				MyBitIndex_ = FoundIndex + 1;

				return true;
			} else {
				// This indicates that iterator has reached the end
				// Caveat: MyArrayIndex must be able to hold NUM_ELEMS
				MyBitIndex_ = NUM_ELEMS*BITS_PER_ELEM;
				return false;
			}
		}

	private:
		BitsetType *  	MyBitset_;
		BitIndexType		MyBitIndex_;
	};

	inline IndexIteratorOnes
	GetIndexIteratorOnes() {
		return IndexIteratorOnes(this);
	}

protected:
	ElemType RawSet_[NUM_ELEMS];

private:
	static inline __attribute__((always_inline)) ArrayIndexType
	BitIndexToArrayIndex(BitIndexType BitIndex) {
		return BitIndex >> BIT_TO_ARRAY_SHIFT;
	}

	// For now we just loop, maybe we trust in loop unrolling or maybe we should use a metaprog later.
	ArrayIndexType
	GetFirstNotZeroElem() const {
		ArrayIndexType i = 0;
		// ATTENTION: As NUM_ELEMS can always be the maximum
		// for ArrayIndexType, we always stop at the second
		// last iteration.
		for( ; i < NUM_ELEMS - 1; ++i) {
			if(RawSet_[i]) {
				break;
			}
		}
		return i;
	}

	ArrayIndexType
	GetFirstNotAllSetElem() const {
		ArrayIndexType i = 0;
		// ATTENTION: As NUM_ELEMS can always be the maximum
		// for ArrayIndexType, we always stop at the second
		// last iteration.
		for( ; i < NUM_ELEMS - 1; ++i) {
			if(RawSet_[i] != ~(ElemType)0) {
				break;
			}
		}
		return i;
	}

	ArrayIndexType
	GetLastNotZeroElem() const {
		ArrayIndexType i = NUM_ELEMS-1;
		// ATTENTION: ArrayIndexType should be unsigned we stop iterations before
		// coming to 0
		for( ; i > 0; --i) {
			if(RawSet_[i]) {
				break;
			}
		}
		return i;
	}

	ArrayIndexType
	GetLastNotAllSetElem() const {
		ArrayIndexType i = NUM_ELEMS-1;
		// ATTENTION: ArrayIndexType should be unsigned we stop iterations before
		// coming to 0
		for( ; i > 0; --i) {
			if(RawSet_[i] != ~(ElemType)0) {
				break;
			}
		}

		return i;
	}
};

template<class ELEM_TYPE, int BIT_LENGTH>
class BitsetMultiElemImpl : public BitsetMultiElemImplBase<ELEM_TYPE, (BIT_LENGTH/(8*sizeof(ELEM_TYPE))) + ((BIT_LENGTH%(8*sizeof(ELEM_TYPE)))?1:0)> {
private:
	// Make dependent of NUM_ELEMS
	enum { LAST_ELEM_BIT_LENGTH = ((BIT_LENGTH-1) % (8*sizeof(ELEM_TYPE))) + 1};
	enum { NUM_ELEMS						= (BIT_LENGTH/(8*sizeof(ELEM_TYPE))) + ((BIT_LENGTH%(8*sizeof(ELEM_TYPE)))?1:0) };

	typedef BitsetMultiElemImplBase<ELEM_TYPE,NUM_ELEMS> BitsetImplBase;

	typedef ELEM_TYPE ElemType;
	typedef typename BitsetImplBase::ArrayIndexType ArrayIndexType;

	static const ElemType SetAllLastElem = metaprog::TwoPowI_minus1<LAST_ELEM_BIT_LENGTH>::RES;

public:
	typedef typename BitsetImplBase::BitIndexType BitIndexType;

	__attribute__((noinline)) void
	SetAll() {
		// FIXME For now simple impl, later maybe metaprog
		ArrayIndexType i = 0;
		for( ; i < NUM_ELEMS-1; ++i) {
			this->RawSet_[i] = ~(ElemType)0;
		}
		this->RawSet_[i] = SetAllLastElem;
	}

	__attribute__((noinline)) Bool
	AllSet() const {
		// FIXME For now simple impl, later maybe metaprog
		ArrayIndexType i = 0;
		ElemType AndedAcc = ~0;
		for( ; i < NUM_ELEMS-1; ++i) {
			AndedAcc &= this->RawSet_[i];
		}
		return (AndedAcc == ~(ElemType)0) && (this->RawSet_[i] == SetAllLastElem);
	}
};

#endif
