#include "../cfInitializers.h"
#include "os/com/ComStack_Types.h"

#include "os/com/ascom/Cfg.h"
#include "os/com/ascom/Filter.h"
#include "os/com/ascom/ConfigTypes.h"

#include "AsCom_Cfg.h"

// It is planned to check the config by templates or directly from inside the configuration script.
// There it should raised an error, if for example a signal has been configured with UInt32, but in p::v config this type
// wasn't configured before. Currently this is not detected and the error will lead to undefined behavior at runtime.

namespace os { namespace com { namespace ascom {

// In the RxSigConfigs normal signals as well as group signals have to be supplied.
// In the example the first 3 signals are associated with RxIpdu 0, the others to RxIpdu 1.
// In the example all signals are values without sign and with different lengths. In the example
// signal 2 has to be extracted from the sdu from byte 3 bit 0 to byte 6 bit 7.
// The signals aren't to be filtered, so there is supplied 0 for the parameters FilterX and FilterY.
// Group signals are generally subject to the restriction of not being filtered. This is accommodated by a special macro.
static RxSignalConfigType const RxSigConfigs [] = {
	OS_COM_ASCOM_RX_SIGNAL_CONFIG_INITIALIZER(0, UInt8,  0, 0, 0, 7, Always, 0, 0),
	OS_COM_ASCOM_RX_SIGNAL_CONFIG_INITIALIZER(0, UInt16, 1, 2, 0, 7, Always, 0, 0),
	OS_COM_ASCOM_RX_SIGNAL_CONFIG_INITIALIZER(0, UInt32, 3, 6, 0, 7, Always, 0, 0),
	OS_COM_ASCOM_RX_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt8, 0, 0, 0, 7),
	OS_COM_ASCOM_RX_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt8, 1, 1, 0, 7),
	OS_COM_ASCOM_RX_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt16, 2, 3, 0, 7),
	OS_COM_ASCOM_RX_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt16, 4, 5, 0, 7)
};

// SignalGroups point to parts of the signal arrays that contain group signals belonging to one signal group.
// In this example we have two signal groups (both are attached to RxIpdu 1),
// one consists of tx group signal 3 - 4 and the other one consists of 5 - 6.
static RxSignalGroupConfigType const RxSignalGroupConfigs [] = {
	OS_COM_ASCOM_RX_SIGNAL_GROUP_CONFIG_INITIALIZER(1, 3, 4),
	OS_COM_ASCOM_RX_SIGNAL_GROUP_CONFIG_INITIALIZER(1, 5, 6)
};

// In this example both RxIpdus belong to Ipdu group 0, they are 7 and 6 bytes long.
// There is no individual IpduCallout and no deadline monitoring configured.
// RxIpdu 0 has associated (non group) signals 0 - 2 and no signal groups.
// RxIpdu 1 has no associated (normal) signals but the signal groups 0 - 1.
static RxIpduConfigType const RxIpduConfigs[] = {
	OS_COM_ASCOM_RX_IPDU_CONFIG_INITIALIZER(0, 0, 2, 1, 0, 7, NULL, 0),
	OS_COM_ASCOM_RX_IPDU_CONFIG_INITIALIZER(0, 1, 0, 0, 1, 6, NULL, 0)
};

// There is one RxIpdu-Group. It is associated with RxIpdus 0 - 1.
static RxIpduGroupConfigType const RxIpduGroupConfigs [] = {
	OS_COM_ASCOM_RX_IPDU_GROUP_CONFIG_INITIALIZER(0, 1)
};

// See RxSignalConfigs.
// Differences:
// - the is an additional parameter TransferProperty which can be TRIGGERED or PENDING.
// - the last Group Signal of a Signal Group must be configured with a special macro. The parameters that are associated with the signal
// group must be supplied in this macro.
static TxSignalConfigType const TxSigConfigs [] = {
	OS_COM_ASCOM_TX_SIGNAL_CONFIG_INITIALIZER(0, UInt8,  0, 0, 0, 7, TRIGGERED, Always, 0, 0),
	OS_COM_ASCOM_TX_SIGNAL_CONFIG_INITIALIZER(0, UInt16, 1, 2, 0, 7, TRIGGERED, Always, 0, 0),
	OS_COM_ASCOM_TX_SIGNAL_CONFIG_INITIALIZER(0, UInt32, 3, 6, 0, 7, TRIGGERED, Always, 0, 0),
	OS_COM_ASCOM_TX_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt8, 0, 0, 0, 7),
	OS_COM_ASCOM_TX_LAST_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt8, 1, 1, 0, 7, TRIGGERED),
	OS_COM_ASCOM_TX_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt16, 2, 3, 0, 7),
	OS_COM_ASCOM_TX_LAST_GROUP_SIGNAL_CONFIG_INITIALIZER(1, UInt16, 4, 5, 0, 7, TRIGGERED)
};

// SignalGroups point to parts of the signal arrays that contain group signals belonging to one signal group.
// In this example we have two signal groups (both are attached to TxIpdu 1),
// one consists of tx group signal 3 - 4 and the other one consists of 5 - 6.
static TxSignalGroupConfigType const TxSignalGroupConfigs [] = {
	OS_COM_ASCOM_TX_SIGNAL_GROUP_CONFIG_INITIALIZER(1, 3, 4),
	OS_COM_ASCOM_TX_SIGNAL_GROUP_CONFIG_INITIALIZER(1, 5, 6)
};

// In this example both TxIpdus belong to Ipdu group 0, signals 0 - 2 and no signal group belong to ipdu 0 no signals and
// signal group 0 - 1 belong to ipdu 1, they are 7 and 6 bytes long.
// There is no individual IpduCallout, minimum delay time and no deadline monitoring configured.
// "false" means there is not transmission mode selection configured.
// Both transmission modes are DIRECT and therefore are configured without periods.
static TxIpduConfigType const TxIpduConfigs[] = {
	OS_COM_ASCOM_TX_IPDU_CONFIG_INITIALIZER(0, 0, 2, 1, 0, 7, NULL, 0, 0, 0, false, DIRECT, 0, 0, DIRECT, 0, 0),
	OS_COM_ASCOM_TX_IPDU_CONFIG_INITIALIZER(0, 1, 0, 0, 1, 6, NULL, 0, 0, 0, false, DIRECT, 0, 0, DIRECT, 0, 0)
};

// There is one TxIpdu-Group. It is associated with TxIpdus 0 - 1.
static TxIpduGroupConfigType const TxIpduGroupConfigs [] = {
	OS_COM_ASCOM_TX_IPDU_GROUP_CONFIG_INITIALIZER(0, 1)
};

extern ConfigType const DefaultConfig;
ConfigType const DefaultConfig = OS_COM_ASCOM_CONFIG_INITIALIZER(RxSigConfigs, 7, RxSignalGroupConfigs, 2, RxIpduConfigs, 2, RxIpduGroupConfigs, 1, TxSigConfigs, 7, TxSignalGroupConfigs, 2, TxIpduConfigs, 2, TxIpduGroupConfigs, 1);

}}}
