/* $Id: dummy.cc,v 1.6 2003/03/26 14:24:01 lauterba Exp $ */

//#define DEBUG
#define FAST
//#define BOGOMIPS 786.43 should be defined on the compiler command line!

// Programmer for the AVR with the "no parts" parallelport adapter 
// PIN LAYOUT ON PARPORT IF:
// 16 - RST
// 11 - MISO
//  2 - MOSI
//  1 - SCK
// 20 - GND
//

#include <sys/io.h>
#include <sched.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

unsigned flashsize = 0x10000;
unsigned pagesize = 256;
unsigned delay_count = (unsigned)(1000000.0 / BOGOMIPS);


struct pin { 
	unsigned char mask; unsigned char port; bool invert; 
	pin(unsigned char m = 0,unsigned char p = 0, bool i = false): invert(i), port(p), mask(m) {}
};

// UISP PINOUT

// The class represents an abstraction for doing SPI communication
// based on parallel port like interface, i.e. the SPI
// protocol is implemented in software
class ParportSPI
{
protected:
	unsigned char last_value[3];
	int base;
public:
	pin SCK; 
	pin DOUT;
	pin DIN; 

	ParportSPI() 
	{
		bzero(last_value,sizeof(last_value));
	}

	// SPI pin lowlevel methods 
	void set(pin& which);
	void clr(pin& which);
	unsigned char get(pin& which);

	// SPI protocol methods 
	unsigned char send_recv(unsigned char);
	void send(unsigned char *data, int size);
	void pulse_sck();
};

// The class represents an abstraction for avr programmers
// based on parallel port like interface, i.e. the SPI
// protocol is implemented in software
class Programmer: public ParportSPI 
{

public:
	pin RESET; 

public:
	


	// AVR initialization for programming methods
	void avr_reset();
	void avr_open();
	void avr_close();
	int avr_enable_programming();


	// AVR programming methods
	unsigned char avr_getpart(unsigned char addr);

	unsigned char avr_read_eeprom(unsigned addr);
	void avr_write_eeprom(unsigned addr, unsigned char byte);

	void avr_prog_flash_byte(unsigned addr, unsigned char byte);
	unsigned char avr_read_flash(unsigned addr); 

	void avr_load_flash(unsigned offset, unsigned char byte);
	void avr_prog_flash(unsigned page);

	int avr_write_flash_byte(unsigned int offset, unsigned char data);
	int avr_write_flash(unsigned page, unsigned char* buf, unsigned size);

	void avr_chip_erase();
	int avr_check_empty();

	// these methods have to be implemented
 	// in the sub classes which realize  specific
	// programmer support
	virtual void adapter_enable() {}
	virtual void adapter_disable() {}
	virtual void isp_enable() {}
	virtual void isp_disable() {}
};

class Dummy: public Programmer
{
public:
	Dummy(int base = 0x378);
};

// Status codes
#define FAILED 1


static void big_delay()
{
#ifndef FAST
	usleep (10);
#else
	// a little delay
	volatile int C = 0;
	for (int i = 0; i < delay_count; i++)
	    C++;
#endif
}

Dummy::Dummy(int nparbase)
{
	base = nparbase;
	SCK = pin( 0x01, 2, true  ); // STROBE inverted
	DOUT= pin( 0x01, 0, false ); // DATA0
	DIN = pin( 0x80, 1, true  );  // BUSY inverted
	RESET = pin( 0x04, 2, false );// INIT
}


class STK300: public Programmer
{
	pin EN1;	
	pin EN2;
	pin LED;
public:
	STK300(int base = 0x378);
	// these methods have to be implemented
 	// in the sub classes which realize  specific
	// programmer support
	void adapter_enable();
	void adapter_disable();
	void isp_enable();
	void isp_disable();
};

STK300::STK300(int nparbase)
{
	base = nparbase;
	SCK = pin(0x10, 0, false  ); // DATA4
	DOUT = pin(0x20, 0, false ); // DATA5
	DIN = pin(0x40, 1, false  );  // ACK  
	RESET= pin(0x80, 0, false ); // DATA7

	EN1 = pin(0x04, 0, false ); // DATA2
	EN2 = pin(0x08, 0, false ); // DATA3
	LED = pin(0x40, 0, false ); // DATA6
}

void STK300::adapter_enable()
{
	clr(EN1);
	clr(EN2);
}
void STK300::adapter_disable()
{

	set(EN1);
	set(EN2);
}
void STK300::isp_enable()
{
	clr(LED);		
}
void STK300::isp_disable()
{
	set(LED);
}


void show_pin(const char* s,pin& which)
{
#if 0
	printf("%s: pin(0x%02x,0x%02x,%d)\n",s,which.mask,which.port,which.invert);
#endif
}


void ParportSPI::set(pin& which)
{
	show_pin("set", which);
	if (!which.invert)
	{
		last_value[which.port] |= which.mask;
	}
	else
	{
		last_value[which.port] &= ~which.mask;
	}
	outb(last_value[which.port], base + which.port);
	big_delay();
}

void ParportSPI::clr(pin& which)
{
	show_pin("clr", which);
	if (which.invert)
	{
		last_value[which.port] |= which.mask;
	}
	else
	{
		last_value[which.port] &= ~which.mask;
	}
	outb(last_value[which.port], base+which.port);
	big_delay();
}

unsigned char ParportSPI::get(pin& which)
{
	show_pin("get", which);
	return which.invert?~inb(base+which.port):inb(base+which.port);
}

unsigned char ParportSPI::send_recv(unsigned char b)
{
	int i, mask, received;

	mask = 0x80;
	received = 0;
	for (i = 0; i < 8; i++) {
		if (b & mask)
			set(DOUT);
		else
			clr(DOUT);
		set(SCK);
		if ( get(DIN) & DIN.mask)
			received |= mask;
		clr(SCK);

		mask >>= 1;
	}
	return received;
}

void ParportSPI::pulse_sck()
{
	clr(SCK);
	set(SCK);
}

void ParportSPI::send(unsigned char *data, int size)
{
	while (size--) {
#ifdef DEBUG
		fprintf(stderr, "%x ", (int)(*data));
#endif
		*data = send_recv(*data);		 	
#ifdef DEBUG
		fprintf(stderr, "(%x) ", (int)(*data));
#endif
		data++;
	}
#ifdef DEBUG
	fprintf(stderr, "\n");
#endif
}


void Programmer::avr_reset()
{
	set(RESET);
	// needed to double this delay on SUSE 9.1. Strange.
	usleep(40000);
	clr(RESET);
	usleep(10000);
}

void Programmer::avr_open() {
	struct sched_param s;
	s.sched_priority = 1;
	sched_setscheduler(0, SCHED_FIFO, &s);
	if (ioperm(base, 8, 1) == 0) {
		adapter_enable();
		isp_enable();		

		// SCK has to be low during the reset
		clr(SCK); 
		
		avr_reset(); // just to be sure give 3 resets
 		avr_reset();
 		avr_reset();
		usleep(20000);
	}

}

void Programmer::avr_close()
{
	avr_reset();
	isp_disable();
	adapter_disable();
}


int Programmer::avr_enable_programming()
{
	int tries = 33;
	do {
		unsigned char cmd[4] = {0xAC,0x53,0,0};
		send(cmd, 4);
		if (cmd[2] == 0x53) return 0;
		pulse_sck();
	} while (tries--);
	return 1;
}
	
unsigned char Programmer::avr_getpart(unsigned char addr)
{
	unsigned char cmd[4] = {0x30,0,addr,0};
	send(cmd,4);
	return cmd[3];
}

unsigned char Programmer::avr_read_eeprom(unsigned addr) 
{
	unsigned char cmd[4] = {0xA0, addr>>8, addr, 0};
	send(cmd,4);
	return cmd[3];
}

void Programmer::avr_write_eeprom(unsigned addr, unsigned char byte) 
{
	unsigned char cmd[4] = {0xC0, addr>>8, addr, byte};
	send(cmd,4);
#ifdef FAST
	usleep(4000);
#else
	usleep(9000);
#endif
}

void Programmer::avr_prog_flash_byte(unsigned addr, unsigned char byte) 
{
	unsigned char cmd[4] = {(addr & 1) ? 0x48 : 0x40,addr>>9,addr>>1, byte};
	send(cmd,4);
#ifdef FAST
	usleep(4000);
#else
	usleep(9000);
#endif
}
unsigned char Programmer::avr_read_flash(unsigned addr) 
{
	unsigned char cmd[4] = {(addr & 1) ? 0x28 : 0x20,addr>>9,addr>>1, 0};
	send(cmd,4);
        return cmd[3]; 
}

void Programmer::avr_load_flash(unsigned offset, unsigned char byte)
{
	unsigned char cmd[4] = {(offset & 1) ? 0x48 : 0x40,0,offset>>1,byte};
	send(cmd,4);
}

void Programmer::avr_prog_flash(unsigned page)
{
	unsigned char cmd[4] = {0x4C, page>>1, (page&1)<<7, 0};
	send(cmd,4);
#ifdef FAST
	usleep(22000);
#else
	usleep(56000);
#endif
}

int Programmer::avr_write_flash_byte(unsigned int offset, unsigned char data)
{
	avr_prog_flash_byte(offset,data);
	unsigned char rdata = avr_read_flash(offset);
	if (rdata != data) {
		fprintf(stderr, "\nverify failed at %x, input = %x, flash = %x\n",
			offset, data, rdata);
		return 1;
	}
	return 0;
}
int Programmer::avr_write_flash(unsigned page, unsigned char* buf, unsigned size)
{
	unsigned offset;
	
	if (size < pagesize) {
		memset(buf + size, 0xff, pagesize - size);
	}
	for (offset = 0; offset < pagesize; offset++) {
		avr_load_flash(offset, buf[offset]);
	}
	avr_prog_flash(page);
	for (offset = 0; offset < size; offset++) {
		unsigned char data = avr_read_flash((page * pagesize) + offset);
		if (data != buf[offset]) {
			fprintf(stderr, "\nverify failed at %x:%x, input = %x, flash = %x\n",
				page, offset, buf[offset], data);
			return 1;
		}
	}
	return 0;
}

void Programmer::avr_chip_erase()
{
	unsigned char cmd[4] = {0xAC, 0x80, 0x00, 0x00};
	send(cmd,4);
	usleep(56000);
}

int Programmer::avr_check_empty()
{
	return ((avr_read_flash(0x0000) == 0xff) && (avr_read_flash(0x0001) == 0xff));
}

void cmd_identify(Programmer* p)
{
	fprintf(stderr, "Vendor=%x Family=%x Number=%x\n",
	        p->avr_getpart(0), p->avr_getpart(1), p->avr_getpart(2));
}

void cmd_erase(Programmer *p)
{
	fprintf(stderr, "erasing flash memory ...\n");
	p->avr_chip_erase();
}

void cmd_read(Programmer* p)
{
	unsigned a;
	fprintf(stderr, "reading flash memory ");
	for (a = 0; a < flashsize; a++) {
		if ((a % pagesize) == 0) fprintf(stderr, "#");
		fputc(p->avr_read_flash(a), stdout);
	}
	fprintf(stderr, "\n");
}

void cmd_write_byte(Programmer* p)
{
	unsigned n = 0;
	int data;

	if (!p->avr_check_empty()) {
		fprintf(stderr, "flash is not empty, erasing ...\n");
		p->avr_chip_erase();
	}
	fprintf(stderr, "programming flash memory ");
	while ((data = fgetc(stdin)) != -1) {
		if (p->avr_write_flash_byte(n,data)) return;
		if (n%pagesize == 0) fprintf(stderr, "#");
		n++;
	}
	fprintf(stderr, "\n");
}
void cmd_write_page(Programmer* p)
{
        unsigned page = 0;
	unsigned n = 0;
	int data;
	unsigned char buf[pagesize];

	if (!p->avr_check_empty()) {
		fprintf(stderr, "flash is not empty, erasing ...\n");
		p->avr_chip_erase();
	}
	fprintf(stderr, "programming flash memory ");
	while ((data = fgetc(stdin)) != -1) {
		buf[n++] = data;
		if (n >= pagesize) {
			if (p->avr_write_flash(page++, buf, pagesize)) return;
			n = 0;
			fprintf(stderr, "#");
		}
	}
	if (n) p->avr_write_flash(page, buf, n);
	fprintf(stderr, "\n");
}

void cmd_verify(Programmer* p)
{
	unsigned a = 0;
	int data;
	fprintf(stderr, "verifying flash memory ...\n");
	while ((data = fgetc(stdin)) != -1) {
		int d1 = p->avr_read_flash(a);
		int d2 = data;
		if (d1 != d2) {
			fprintf(stderr, "difference at %x input: %x flash: %x\n",
				a, (unsigned)d2, (unsigned)d1);
			return;
		}
		a++;
	}
}

int main(int argc, char** argv)
{

	Dummy dummy_p;
	STK300 stk300_p;
	Programmer* programmer= &dummy_p;

	if (argc < 2) {
		printf( "Dumb AVR programmer for Linux\n"
		        "Usage: dummy command [configuration]\n"
			"Commands:\n"
			"   [i]dentify      dump chip identification\n"
			"	[e]rase		    erase chip\n"
			"	[r]ead		    read program memory to stdout\n"
			"	[w]rite		    program memory from stdin(paged -> Mega)\n"
			"	[W]rite		    program memory from stdin(byte  -> other AVRs)\n"
			"   write ee[p]rom  program eeprom from stdin\n"
			"	[v]erify        verify program memory with stdin\n"
			"Configurations:     (at most one of the options!)\n"
			"	[n]oinvert  	uninverted i/o for all prog pins (default)\n"
			"	[i]nvert  	    inverted MOSI signal\n"
			"	[S]TK300  	    STK300/STK200 programmer connected\n"
			
			);			
	} else {
		char *what = argv[1];
		char *which = "n";

		if (argc == 3)
		{
			which = argv[2];
		}

		switch (*which) {
			case 'i': dummy_p.DOUT.invert = true; break;
			case 'S': programmer = &stk300_p; break;
			case 'n':
			default : dummy_p.DOUT.invert = false; break;
		}

		programmer->avr_open();
		if (programmer->avr_enable_programming() == FAILED) {
			fprintf(stderr, "dummy: failed to enable programming\n");
			programmer->avr_close();
			return -1;
		}
		programmer->avr_getpart(0);
		programmer->avr_getpart(1);
		programmer->avr_getpart(2);
		switch (*what) {
			case 'i': cmd_identify(programmer); break;
			case 'e': cmd_erase(programmer); break;
			case 'r': cmd_read(programmer); break;
			case 'W': cmd_write_byte(programmer); break;
			case 'w': cmd_write_page(programmer); break;
			case 'v': cmd_verify(programmer); break;
		}
		programmer->avr_close();
	}
	return 0;
}
