#!/usr/bin/perl

use warnings;
use strict;

use SystemConfiguration;

if(@ARGV != 3 || $ARGV[0] !~ /\.xml$/ || $ARGV[2] !~ /\.cpp$/) {
	print STDERR "Usage: $0 <config.xml> <os.config> <main.cpp>\n";
	exit 1;
}

my $cfg = SystemConfiguration->new($ARGV[0], $ARGV[1]);

## Here comes the actual generation code
open (SRC, ">$ARGV[2]") || die "cannot open $ARGV[2]\n";

print SRC <<EOH;
// generated CiAO runtime
#include "hw/hal/MPU.h"
#include "as/autosar_types.h"
#include "as/AS.h"
#include "../cfInitializers.h"
#include "objects.h"

// usercode includes
EOH


print SRC "#include \"$_\"\n" foreach $cfg->getAppHeaders();

if($cfg->hasOSFeature('timeprot')) {
	print SRC <<EOH;
//include header of system clock for time specifications for timing protection
#include "hw/dev/STM.h"

EOH
}

# Linker Symbols
print SRC <<EOS;

// symbols
extern char __CIAO_NTCODE_BEGIN;
extern char __CIAO_NTCODE_END;
extern char __CIAO_NTDATA_BEGIN;
extern char __CIAO_NTDATA_END;
extern char __CIAO_TCODE_BEGIN;
extern char __CIAO_TCODE_END;
extern char __CIAO_TDATA_BEGIN;
extern char __CIAO_TDATA_END;
extern char __CIAO_OSCODE_BEGIN;
extern char __CIAO_OSCODE_END;
extern char __CIAO_OSDATA_BEGIN;
extern char __CIAO_OSDATA_END;
EOS

foreach my $app ($cfg->getApplications()) {
	foreach my $sym (qw(CODE_BEGIN CODE_END DATA_BEGIN DATA_END STACKS_BEGIN)) {
		print SRC "extern char __CIAO_".$app->{name}."$sym;\n"
	}
}

foreach my $task ($cfg->getTasks()) {
	print SRC "extern char __CIAO_".$task->{name}."STACK_BEGIN;\n";
	print SRC "extern char __CIAO_".$task->{name}."STACK_END;\n";
}

# Stacks
print SRC "\nnamespace os { namespace krn {\n";
foreach my $task ($cfg->getTasks()) {
	print SRC "const UInt $task->{name}_stackelements = ($task->{stacksize}-1)/sizeof(UInt32)+1;\n";
	print SRC "UInt $task->{name}_stack[$task->{name}_stackelements];\n";
}
print SRC "}}\n\n";

# Tasks
print SRC <<EOT;

#include "os/krn/Task.h"
namespace os { namespace krn {
Task theTasks[] = {
EOT

foreach my $t ($cfg->getTasks()) {
	print SRC "TASK_INITIALIZER($t->{priority}, TASK_SUSPENDED, ".$t->entryFunction();
	print SRC ", &$t->{name}_stack[$t->{name}_stackelements],";
	print SRC "0, $t->{component}->{osapp}->{name}, 0, $t->{priority}, ";
	print SRC "(unsigned) sizeof($t->{name}_stack), 0, $t->{activation}, 0, 0, ";
	# preemptable
	print SRC ($t->{schedule} eq 'full' ? '1' : '0');
	# current application
	print SRC ", $t->{component}->{osapp}->{name}, ";
	# upper stack protection bound
	print SRC "&$t->{name}_stack[$t->{name}_stackelements],";
	# bottom of stack
	print SRC "$t->{name}_stack, ";
	# number of domain transitions
	print SRC "0, ";
	# ACL (head pointer)
	print SRC "0, 0, ";
	# inter arrival rate, AR_allowed
	print SRC (exists $t->{timeframe} ? $t->{timeframe} : '0')." * (hw::dev::STM::FREQ / 1000000), true, ";
	# execution time budget
	print SRC (exists $t->{executionbudget} ? $t->{executionbudget} : '0')." * (hw::dev::STM::FREQ / 1000000), ";
	# execution deadline
	print SRC "0, ".(exists $t->{executionbudget} ? $t->{executionbudget} : '0')." * (hw::dev::STM::FREQ / 1000000), ";
	# locking time budget
	print SRC (exists $t->{isrlockbudget} ? $t->{isrlockbudget} : '0')." * (hw::dev::STM::FREQ / 1000000), 0, 0, ";
	# resource locking
	print SRC (exists $t->{resourcelockbudget} ? $t->{resourcelockbudget} : '0')." * (hw::dev::STM::FREQ / 1000000), 0, 0),\n";
}

print SRC "};\n}}\n";

# Alarms
if($cfg->hasOSFeature('kernel_alarms')) {
	# theCounters
	print SRC <<EOAL;
#include "os/krn/Counter.h"
#include "os/krn/Alarm.h"

namespace os { namespace krn {
Counter theCounters[] = {
EOAL

	my $counter_id = 0;
	foreach my $a ($cfg->getCounters()) {
		$a->{id} = $counter_id++;
		print SRC "{ $a->{maxallowedvalue}, $a->{ticksperbase}, $a->{mincycle}, 0 },\n";
	}

	print SRC "};\n\n";

	# theAlarms
	print SRC <<EOAL;
Alarm theAlarms[] = {
EOAL

	foreach my $a ($cfg->getAlarms()) {
		print SRC "{ theCounters + ".$a->getCounter()->{id}.", $a->{armed}, $a->{absolutetime}, $a->{cycletime}, $a->{task}";
		if($cfg->hasOSFeature('kernel_events')) {
			print SRC ", " . ( (exists($a->{setevent}) && $a->{setevent} eq 'true') ? $a->{event} : '0' );
		}
 		print SRC "	},\n";
	}

	print SRC "};\n}}\n\n";
}

# ISRs
if(scalar($cfg->getISRs()) > 0) {
	print SRC '//include device header files for access to their hw::dev::IRQCtrl::reg() functions'."\n";
	foreach my $isr ($cfg->getISRs()) {
		print SRC "#include \"hw/irq/$isr->{device}.h\"\n";
	}

	print SRC <<EOI;
#include "os/krn/ISR.h"
namespace os { namespace krn {
ISR theISRs[] = {
EOI

	foreach my $isr ($cfg->getISRs()) {
		print SRC "ISR_INITIALIZER(".$isr->entryFunction().", (::hw::hal::ServiceRequestRegisterType)(hw::irq::$isr->{device}::REGISTER), ";
		print SRC $isr->{component}->{osapp}->{name}.", ";
		print SRC (exists $isr->{executionbudget} ? $isr->{executionbudget} : '0').' * (hw::dev::STM::FREQ / 1000000), 0, ';
		print SRC (exists $isr->{timeframe} ? $isr->{timeframe} : '0').' * (hw::dev::STM::FREQ / 1000000), true, ';
		print SRC (exists $isr->{isrlockbudget} ? $isr->{isrlockbudget} : '0').' * (hw::dev::STM::FREQ / 1000000), 0, 0, ';
		print SRC (exists $isr->{resourcelockbudget} ? $isr->{resourcelockbudget} : '0').' * (hw::dev::STM::FREQ / 1000000), 0, 0),';
	}
	print SRC "};\n}}\n\n";
}

# Appmodes
print SRC <<EOAM;

#include "os/krn/AppMode.h"
namespace os { namespace krn {
AppMode theAppModes[] = {
EOAM
foreach my $appmode ($cfg->getAppModes()) {
	my $atinit = join('+', map { "( 1 << $_->{name} )" } $appmode->getTasks());
	$atinit = 0 if $atinit eq '';

	if($cfg->hasOSFeature('kernel_alarms')) {
		my $aainit = join('+', map { "( 1 << $_->{name} )" } $appmode->getAlarms());
		$atinit .= ', '. ($aainit eq '' ? 0 : $aainit);
	}

	print SRC "{ $atinit },\n";
}
print SRC "};\n}}\n\n";

# TODO Resources
if($cfg->hasOSFeature('kernel_resources')) {
	print SRC <<EOR;
#include "os/krn/Resource.h"
namespace os { namespace krn {
Resource theResources[] = {
EOR

	foreach my $r ($cfg->getResources()) {
		print SRC "{ $r->{ceiling} }, ";
	}

	print SRC "};\n}}\n\n";
}

# Applications
if($cfg->hasOSFeature('memprot_deferact')) {
	foreach my $app ($cfg->getApplications()) {
		if($cfg->hasOSFeature('kernel_multipleactivation')) {
			print SRC "os::krn::ActivationsType ".$app->appSymbolPrefix()."deferredActivations[cfOS_NUMBER_OF_TASKS] = { ";
			print SRC join(', ', map { '0' } (1..$cfg->hasOSFeature('cfOS_NUMBER_OF_TASKS')));
			print SRC " };\n";
		} else {
			print SRC "uint_t_value<cfOS_NUMBER_OF_TASKS>::T ".$app->appSymbolPrefix()."deferredActivations = 0;\n";
		}
	}
	print SRC "\n\n";
}
if($cfg->hasOSFeature('memprot_defereve')) {
	foreach my $app ($cfg->getApplications()) {
		print SRC "os::krn::EventMaskType ".$app->appSymbolPrefix()."deferredEvents[cfOS_NUMBER_OF_TASKS] = { ";
		print SRC join(', ', map { '0' } (1..$cfg->hasOSFeature('cfOS_NUMBER_OF_TASKS')));
		print SRC " };\n";
	}
	print SRC "\n\n";
}


print SRC <<EOA;
#include "os/krn/Application.h"
namespace os { namespace krn {
Application theApplications[] = {
EOA
foreach my $app ($cfg->getApplications()) {
	print SRC "APPLICATION_INITIALIZER($app->{trusted}, ";
	foreach my $sym (qw(DATA_BEGIN DATA_END STACKS_BEGIN CODE_BEGIN CODE_END)) {
		print SRC "(Ptr) &__CIAO_$app->{name}$sym, ";
	}
	print SRC $app->appSymbolPrefix()."deferredActivations, ";
	print SRC $app->appSymbolPrefix()."deferredEvents),\n";
}
print SRC "};\n}}\n\n";

# component instances
foreach my $c ($cfg->getComponents()) {
	print SRC "$c->{name} $c->{name}::inst_;\n";
}

# main()
print SRC <<EOM;

int CIAO_WEAK main() {
	AS::StartOS (OSDEFAULTAPPMODE);
	_debug();
	return 0;
}
EOM


close(SRC);
